#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013 Artem Glebov <artem.glebov@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk

# -------------------------------------------------------------------------
#
# grabbers constants and routines
#
# -------------------------------------------------------------------------

GRABBER_INSIDE = 0
GRABBER_OUTSIDE = 1

MIN_CORNER_GRABBER = 20
MIN_SIDE_GRABBER = 20
MIN_GRABBER_PADDING = 10
MIN_SIDE_FOR_INSIDE_GRABBERS = (
    2 * (MIN_CORNER_GRABBER + MIN_GRABBER_PADDING) + MIN_SIDE_GRABBER
)

INSIDE = 0
GRABBER_UPPER_LEFT = 1
GRABBER_UPPER = 2
GRABBER_UPPER_RIGHT = 3
GRABBER_RIGHT = 4
GRABBER_LOWER_RIGHT = 5
GRABBER_LOWER = 6
GRABBER_LOWER_LEFT = 7
GRABBER_LEFT = 8


def upper_left_grabber_inner(x1, y1, x2, y2):
    return (x1, y1, x1 + MIN_CORNER_GRABBER, y1 + MIN_CORNER_GRABBER)


def upper_grabber_inner(x1, y1, x2, y2):
    return (
        x1 + MIN_CORNER_GRABBER + MIN_GRABBER_PADDING,
        y1,
        x2 - MIN_CORNER_GRABBER - MIN_GRABBER_PADDING,
        y1 + MIN_CORNER_GRABBER,
    )


def upper_right_grabber_inner(x1, y1, x2, y2):
    return (x2 - MIN_CORNER_GRABBER, y1, x2, y1 + MIN_CORNER_GRABBER)


def right_grabber_inner(x1, y1, x2, y2):
    return (
        x2 - MIN_CORNER_GRABBER,
        y1 + MIN_CORNER_GRABBER + MIN_GRABBER_PADDING,
        x2,
        y2 - MIN_CORNER_GRABBER - MIN_GRABBER_PADDING,
    )


def lower_right_grabber_inner(x1, y1, x2, y2):
    return (x2 - MIN_CORNER_GRABBER, y2 - MIN_CORNER_GRABBER, x2, y2)


def lower_grabber_inner(x1, y1, x2, y2):
    return (
        x1 + MIN_CORNER_GRABBER + MIN_GRABBER_PADDING,
        y2 - MIN_CORNER_GRABBER,
        x2 - MIN_CORNER_GRABBER - MIN_GRABBER_PADDING,
        y2,
    )


def lower_left_grabber_inner(x1, y1, x2, y2):
    return (x1, y2 - MIN_CORNER_GRABBER, x1 + MIN_CORNER_GRABBER, y2)


def left_grabber_inner(x1, y1, x2, y2):
    return (
        x1,
        y1 + MIN_CORNER_GRABBER + MIN_GRABBER_PADDING,
        x1 + MIN_CORNER_GRABBER,
        y2 - MIN_CORNER_GRABBER - MIN_GRABBER_PADDING,
    )


# outer


def upper_left_grabber_outer(x1, y1, x2, y2):
    return (x1 - MIN_CORNER_GRABBER, y1 - MIN_CORNER_GRABBER, x1, y1)


def upper_grabber_outer(x1, y1, x2, y2):
    return (x1, y1 - MIN_CORNER_GRABBER, x2, y1)


def upper_right_grabber_outer(x1, y1, x2, y2):
    return (x2, y1 - MIN_CORNER_GRABBER, x2 + MIN_CORNER_GRABBER, y1)


def right_grabber_outer(x1, y1, x2, y2):
    return (x2, y1, x2 + MIN_CORNER_GRABBER, y2)


def lower_right_grabber_outer(x1, y1, x2, y2):
    return (x2, y2, x2 + MIN_CORNER_GRABBER, y2 + MIN_CORNER_GRABBER)


def lower_grabber_outer(x1, y1, x2, y2):
    return (x1, y2, x2, y2 + MIN_CORNER_GRABBER)


def lower_left_grabber_outer(x1, y1, x2, y2):
    return (x1 - MIN_CORNER_GRABBER, y2, x1, y2 + MIN_CORNER_GRABBER)


def left_grabber_outer(x1, y1, x2, y2):
    return (x1 - MIN_CORNER_GRABBER, y1, x1, y2)


# motion


def inside_moved(x1, y1, x2, y2, dx, dy):
    return (x1, y1, x2, y2)


def upper_left_moved(x1, y1, x2, y2, dx, dy):
    return (x1 + dx, y1 + dy, x2, y2)


def upper_moved(x1, y1, x2, y2, dx, dy):
    return (x1, y1 + dy, x2, y2)


def upper_right_moved(x1, y1, x2, y2, dx, dy):
    return (x1, y1 + dy, x2 + dx, y2)


def right_moved(x1, y1, x2, y2, dx, dy):
    return (x1, y1, x2 + dx, y2)


def lower_right_moved(x1, y1, x2, y2, dx, dy):
    return (x1, y1, x2 + dx, y2 + dy)


def lower_moved(x1, y1, x2, y2, dx, dy):
    return (x1, y1, x2, y2 + dy)


def lower_left_moved(x1, y1, x2, y2, dx, dy):
    return (x1 + dx, y1, x2, y2 + dy)


def left_moved(x1, y1, x2, y2, dx, dy):
    return (x1 + dx, y1, x2, y2)


# switching

GRABBERS = [
    INSIDE,
    GRABBER_UPPER_LEFT,
    GRABBER_UPPER,
    GRABBER_UPPER_RIGHT,
    GRABBER_RIGHT,
    GRABBER_LOWER_RIGHT,
    GRABBER_LOWER,
    GRABBER_LOWER_LEFT,
    GRABBER_LEFT,
]

INNER_GRABBERS = [
    None,
    upper_left_grabber_inner,
    upper_grabber_inner,
    upper_right_grabber_inner,
    right_grabber_inner,
    lower_right_grabber_inner,
    lower_grabber_inner,
    lower_left_grabber_inner,
    left_grabber_inner,
]

OUTER_GRABBERS = [
    None,
    upper_left_grabber_outer,
    upper_grabber_outer,
    upper_right_grabber_outer,
    right_grabber_outer,
    lower_right_grabber_outer,
    lower_grabber_outer,
    lower_left_grabber_outer,
    left_grabber_outer,
]

MOTION_FUNCTIONS = [
    inside_moved,
    upper_left_moved,
    upper_moved,
    upper_right_moved,
    right_moved,
    lower_right_moved,
    lower_moved,
    lower_left_moved,
    left_moved,
]

GRABBERS_SWITCH = [
    [INSIDE, INSIDE, INSIDE],
    [GRABBER_UPPER_RIGHT, GRABBER_LOWER_RIGHT, GRABBER_LOWER_LEFT],
    [GRABBER_UPPER, GRABBER_LOWER, GRABBER_LOWER],
    [GRABBER_UPPER_LEFT, GRABBER_LOWER_LEFT, GRABBER_LOWER_RIGHT],
    [GRABBER_LEFT, GRABBER_LEFT, GRABBER_RIGHT],
    [GRABBER_LOWER_LEFT, GRABBER_UPPER_LEFT, GRABBER_UPPER_RIGHT],
    [GRABBER_LOWER, GRABBER_UPPER, GRABBER_UPPER],
    [GRABBER_LOWER_RIGHT, GRABBER_UPPER_RIGHT, GRABBER_UPPER_LEFT],
    [GRABBER_RIGHT, GRABBER_RIGHT, GRABBER_LEFT],
]

# cursors
try:
    CURSOR_UPPER = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.TOP_SIDE
    )
    CURSOR_LOWER = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.BOTTOM_SIDE
    )
    CURSOR_LEFT = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.LEFT_SIDE
    )
    CURSOR_RIGHT = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.RIGHT_SIDE
    )
    CURSOR_UPPER_LEFT = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.TOP_LEFT_CORNER
    )
    CURSOR_UPPER_RIGHT = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.TOP_RIGHT_CORNER
    )
    CURSOR_LOWER_LEFT = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.BOTTOM_LEFT_CORNER
    )
    CURSOR_LOWER_RIGHT = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.BOTTOM_RIGHT_CORNER
    )
except:
    CURSOR_UPPER = None
    CURSOR_LOWER = None
    CURSOR_LEFT = None
    CURSOR_RIGHT = None
    CURSOR_UPPER_LEFT = None
    CURSOR_UPPER_RIGHT = None
    CURSOR_LOWER_LEFT = None
    CURSOR_LOWER_RIGHT = None

CURSORS = [
    None,
    CURSOR_UPPER_LEFT,
    CURSOR_UPPER,
    CURSOR_UPPER_RIGHT,
    CURSOR_RIGHT,
    CURSOR_LOWER_RIGHT,
    CURSOR_LOWER,
    CURSOR_LOWER_LEFT,
    CURSOR_LEFT,
]

# helper functions


def grabber_position(rect):
    x1, y1, x2, y2 = rect
    if (
        x2 - x1 >= MIN_SIDE_FOR_INSIDE_GRABBERS
        and y2 - y1 >= MIN_SIDE_FOR_INSIDE_GRABBERS
    ):
        return GRABBER_INSIDE
    else:
        return GRABBER_OUTSIDE


def grabber_generators(rect):
    if grabber_position(rect) == GRABBER_INSIDE:
        return INNER_GRABBERS
    else:
        return OUTER_GRABBERS


def switch_grabber(grabber, x1, y1, x2, y2):
    switch_row = GRABBERS_SWITCH[grabber]
    if x1 > x2:
        if y1 > y2:
            return switch_row[1]
        else:
            return switch_row[0]
    else:
        if y1 > y2:
            return switch_row[2]
        else:
            return grabber


def inside_rect(rect, x, y):
    x1, y1, x2, y2 = rect
    return x1 <= x <= x2 and y1 <= y <= y2


def can_grab(rect, x, y):
    """
    Checks if (x,y) lies within one of the grabbers of rect.
    """
    (x1, y1, x2, y2) = rect
    if (
        x2 - x1 >= MIN_SIDE_FOR_INSIDE_GRABBERS
        and y2 - y1 >= MIN_SIDE_FOR_INSIDE_GRABBERS
    ):
        # grabbers are inside
        if x < x1 or x > x2 or y < y1 or y > y2:
            return None
        for grabber in GRABBERS[1:]:
            grabber_area = INNER_GRABBERS[grabber](x1, y1, x2, y2)
            if inside_rect(grabber_area, x, y):
                return grabber
        return INSIDE
    else:
        # grabbers are outside
        if x1 <= x <= x2 and y1 <= y <= y2:
            return INSIDE
        for grabber in GRABBERS[1:]:
            grabber_area = OUTER_GRABBERS[grabber](x1, y1, x2, y2)
            if inside_rect(grabber_area, x, y):
                return grabber
        return None



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013       John Ralls <jralls@ceridwen.us>
# Copyright (C) 2013-2017  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Commonly used report options. Call the function, don't copy the code!
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...config import config
from ...datehandler import get_date_formats, LANG_TO_DISPLAY, main_locale
from ...display.name import displayer as global_name_display
from ...lib.date import Today
from ...display.place import displayer as _pd
from ..menu import EnumeratedListOption, BooleanOption, NumberOption
from ...proxy import PrivateProxyDb, LivingProxyDb
from ...utils.grampslocale import GrampsLocale
from ...const import GRAMPS_LOCALE as glocale
from ...utils.place import coord_formats, coord_formats_desc

_ = glocale.translation.sgettext


# _T_ is a gramps-defined keyword -- see po/update_po.py and po/genpot.sh
def _T_(value, context=""):  # enable deferred translations
    return "%s\x04%s" % (context, value) if context else value


# -------------------------------------------------------------------------
#
# StandardReportOptions
#
# -------------------------------------------------------------------------


def add_localization_option(menu, category):
    """
    Insert an option for localizing the report into a different locale
    from the UI locale.
    """

    trans = EnumeratedListOption(_("Translation"), glocale.DEFAULT_TRANSLATION_STR)
    trans.add_item(glocale.DEFAULT_TRANSLATION_STR, _("Default"))
    languages = glocale.get_language_dict()
    for language in sorted(languages, key=glocale.sort_key):
        trans.add_item(languages[language], language)
    trans.set_help(_("The translation to be used for the report."))
    menu.add_option(category, "trans", trans)
    return trans


def add_extra_localization_option(menu, category, name, optname):
    """
    Insert an option for localizing the report into a different locale
    than the default one
    """

    trans = EnumeratedListOption(_(name), glocale.DEFAULT_TRANSLATION_STR)
    trans.add_item(glocale.DEFAULT_TRANSLATION_STR, _("Default"))
    languages = glocale.get_language_dict()
    for language in sorted(languages, key=glocale.sort_key):
        trans.add_item(languages[language], language)
    trans.set_help(_("The additional translation to be used for the report."))
    menu.add_option(category, optname, trans)
    return trans


def add_name_format_option(menu, category):
    """
    Insert an option for changing the report's name format to a
    report-specific format instead of the user's Edit=>Preferences choice
    """
    name_format = EnumeratedListOption(_("Name format"), 0)
    name_format.add_item(0, _("Default"))
    format_list = global_name_display.get_name_format()
    for number, name, format_string, whether_active in format_list:
        name_format.add_item(number, name)
    name_format.set_help(_("Select the format to display names"))
    current_format = config.get("preferences.name-format")
    # if this report hasn't ever been run, start with user's current setting
    name_format.set_value(current_format)
    # if the report has been run, this line will get the user's old setting
    menu.add_option(category, "name_format", name_format)
    return name_format


def run_name_format_option(report, menu):
    """
    Run the option for changing the report's name format to a
    report-specific format instead of the user's Edit=>Preferences choice
    """
    current_format = config.get("preferences.name-format")
    name_format = menu.get_option_by_name("name_format").get_value()
    if name_format != current_format:
        report._name_display.set_default_format(name_format)
    return name_format


def add_private_data_option(menu, category, default=True):
    """
    Insert an option for deciding whether the information in the
    database marked "private" shall be included in the report

    Since historically, before this option, the entire database was
    used, including private information, the default for this option
    has been set to be True, to include such private information.
    """
    incl_private = BooleanOption(_("Include data marked private"), default)
    incl_private.set_help(_("Whether to include private data"))
    menu.add_option(category, "incl_private", incl_private)


def run_private_data_option(report, menu):
    """
    Run the option for deciding whether the information in the
    database marked "private" shall be included in the report
    """
    include_private_data = menu.get_option_by_name("incl_private").get_value()
    if not include_private_data:
        report.database = PrivateProxyDb(report.database)


def add_living_people_option(
    menu,
    category,
    mode=LivingProxyDb.MODE_INCLUDE_ALL,
    after_death_years=0,
    process_names=True,
):
    """
    Insert an option for deciding how the information in the
    database for living people shall be handled by the report

    Because historically, before this option, the entire database was
    used, including all information on all living people, the default
    mode for this option has been set to include all such information

    The value of the "living_people" option is the same as the value
    of the "mode" argument in the call to the LivingProxyDb proxy DB

    :param menu: The menu the options should be added to.
    :type menu: :class:`.Menu`
    :param category: A label that describes the category that the option
        belongs to.
        Example: "Report Options"
    :type category: string
    :param mode:
        The method for handling living people.
        LivingProxyDb.MODE_EXCLUDE_ALL will remove living people altogether.
        LivingProxyDb.MODE_INCLUDE_LAST_NAME_ONLY will remove all
            information and change their given name to "[Living]" or whatever
            has been set in Preferences -> Text -> Private given name.
        LivingProxyDb.MODE_REPLACE_COMPLETE_NAME will remove all
            information and change their given name and surname to
            "[Living]" or whatever has been set in Preferences -> Text
            for Private surname and Private given name.
        LivingProxyDb.MODE_INCLUDE_FULL_NAME_ONLY will remove all
            information but leave the entire name intact.
        LivingProxyDb.MODE_INCLUDE_ALL will not invoke LivingProxyDb at all.
    :type mode: int
    :param after_death_years:
        The number of years after a person's death to
        still consider them as living.
    :type after_death_years: int
    :return: nothing
    :param process_names: whether to offer name-oriented option choices
    :type process_names: Boolean
    """

    def living_people_changed():
        """
        Handle a change in the living_people option
        """
        if living_people.get_value() == LivingProxyDb.MODE_INCLUDE_ALL:
            years_past_death.set_available(False)
        else:
            years_past_death.set_available(True)

    living_people = EnumeratedListOption(_("Living People"), mode)
    items = [
        (
            LivingProxyDb.MODE_INCLUDE_ALL,
            _T_("Included, and all data", "'living people'"),
        )
    ]
    if process_names:
        items += [
            (
                LivingProxyDb.MODE_INCLUDE_FULL_NAME_ONLY,
                _T_("Full names, but data removed", "'living people'"),
            ),
            (
                LivingProxyDb.MODE_INCLUDE_LAST_NAME_ONLY,
                _T_("Given names replaced, and data removed", "'living people'"),
            ),
            (
                LivingProxyDb.MODE_REPLACE_COMPLETE_NAME,
                _T_("Complete names replaced, and data removed", "'living people'"),
            ),
        ]
    items += [(LivingProxyDb.MODE_EXCLUDE_ALL, _T_("Not included", "'living people'"))]
    living_people.set_items(items, xml_items=True)  # for deferred translation
    living_people.set_help(_("How to handle living people"))
    menu.add_option(category, "living_people", living_people)
    living_people.connect("value-changed", living_people_changed)
    years_past_death = NumberOption(
        _("Years from death to consider living"), after_death_years, 0, 100
    )
    years_past_death.set_help(_("Whether to restrict data on recently-dead people"))
    menu.add_option(category, "years_past_death", years_past_death)
    living_people_changed()


def run_living_people_option(report, menu, llocale=glocale):
    """
    Run the option for deciding how the information in the
    database for living people shall be handled by the report

    If llocale is passed in (a :class:`.GrampsLocale`), then (insofar as
    possible) the translated values will be returned instead.

    :param llocale: allow deferred translation of "[Living]"
    :type llocale: a :class:`.GrampsLocale` instance
    """
    option = menu.get_option_by_name("living_people")
    living_value = option.get_value()
    years_past_death = menu.get_option_by_name("years_past_death").get_value()
    if living_value != LivingProxyDb.MODE_INCLUDE_ALL:
        report.database = LivingProxyDb(
            report.database,
            living_value,
            years_after_death=years_past_death,
            llocale=llocale,
        )
    return option


def add_date_format_option(menu, category, localization_option):
    """
    Insert an option for changing the report's date format to a
    report-specific format instead of the user's Edit=>Preferences choice

    :param localization_option: allow realtime translation of date formats
    :type localization_option: a :class:`.EnumeratedListOption` instance
    """

    def on_trans_value_changed():
        """
        Handle a change in the localization option (inside the date-format one)
        """
        lang = localization_option.get_value()
        if lang == GrampsLocale.DEFAULT_TRANSLATION_STR:  # the UI language
            vlocale = glocale
        elif lang in LANG_TO_DISPLAY:  # a displayer exists
            vlocale = LANG_TO_DISPLAY[lang]._locale  # locale is already loaded
        else:  # no displayer
            vlocale = GrampsLocale(lang=main_locale[lang])
        ldd = vlocale.date_displayer
        target_format_list = get_date_formats(vlocale)  # get localized formats
        # trans_text is a defined keyword (see po/update_po.py, po/genpot.sh)
        trans_text = vlocale.translation.sgettext
        if global_date_format < len(target_format_list):
            ldd.set_format(global_date_format)
            example = vlocale.get_date(today)
            target_option_list = [
                (
                    0,
                    "%s (%s) (%s)"
                    % (
                        trans_text("Default"),
                        target_format_list[global_date_format],
                        example,
                    ),
                )
            ]
        else:
            target_option_list = [(0, trans_text("Default"))]
        for fmt in target_format_list:
            index = target_format_list.index(fmt) + 1  # option default = 0
            ldd.set_format(index - 1)
            example = vlocale.get_date(today)
            target_option_list += [(index, fmt + " (%s)" % example)]
        date_option.set_items(target_option_list)

    today = Today()
    date_option = EnumeratedListOption(_("Date format"), 0)
    global_date_format = config.get("preferences.date-format")
    on_trans_value_changed()
    date_option.set_help(_("The format and language for dates, with examples"))
    # if this report hasn't ever been run, start with user's current setting
    date_option.set_value(global_date_format + 1)
    # if the report has been run, this line will get the user's old setting
    menu.add_option(category, "date_format", date_option)
    localization_option.connect("value-changed", on_trans_value_changed)


def run_date_format_option(report, menu):
    """
    Run the option for changing the report's date format to a
    report-specific format instead of the user's Edit=>Preferences choice
    """

    def warning(value):
        """
        Convenience function for the error message.

        The 'value' (starting at '0') is the index in the option choices list
        (as opposed to the index (starting at '0') in the target format list,
        which is shorter by one, so corresponding offsets are made).
        """
        target_format_choices = date_option.get_items()  # with numbers
        report._user.warn(
            _("Ignoring unknown option: %s") % value,
            _("Valid values: ")
            + str(target_format_choices)
            + "\n\n"
            + _("Using options string: %s") % str(target_format_list[0]),
        )  # the previous line's "0" is because ISO is always the fallback

    target_format_list = get_date_formats(report._locale)
    date_option = menu.get_option_by_name("date_format")
    date_opt_value = date_option.get_value()
    if date_opt_value == 0:  # "default"
        format_to_be = config.get("preferences.date-format")  # the UI choice
    elif date_opt_value <= len(target_format_list):
        format_to_be = date_opt_value - 1
    else:
        warning(date_opt_value)
        format_to_be = 0  # ISO always exists
    if format_to_be + 1 > len(target_format_list):
        warning(format_to_be + 1)
        format_to_be = 0  # ISO always exists
    report._ldd.set_format(format_to_be)


def add_tags_option(menu, category):
    """
    Insert an option for deciding whether to include tags
    in the report

    :param menu: The menu the options should be added to.
    :type menu: :class:`.Menu`
    :param category: A label that describes the category that the option
        belongs to, e.g. "Report Options"
    :type category: string
    """

    include_tags = EnumeratedListOption(_("Tags"), 0)
    include_tags.add_item(0, _("Do not include"))
    include_tags.add_item(1, _("Include"))
    include_tags.set_help(_("Whether to include tags"))
    menu.add_option(category, "inc_tags", include_tags)


def add_gramps_id_option(menu, category, ownline=False):
    """
    Insert an option for deciding whether to include Gramps IDs
    in the report

    Since for some reports it makes sense to possibly have the ID on its
    own line (e.g. Graphviz reports), that possibility is included, but
    since for most reports it won't make sense the default is False

    :param menu: The menu the options should be added to.
    :type menu: :class:`.Menu`
    :param category: A label that describes the category that the option
        belongs to, e.g. "Report Options"
    :type category: string
    :param ownline: whether the option offers to have the ID on its own line
    :type ownline: Boolean
    """

    include_id = EnumeratedListOption(_("Gramps ID"), 0)
    include_id.add_item(0, _("Do not include"))
    if ownline:
        include_id.add_item(1, _("Share an existing line"))
        include_id.add_item(2, _("On a line of its own"))
        include_id.set_help(_("Whether (and where) to include Gramps IDs"))
    else:
        include_id.add_item(1, _("Include"))
        include_id.set_help(_("Whether to include Gramps IDs"))
    menu.add_option(category, "inc_id", include_id)


def add_place_format_option(menu, category):
    """
    Insert an option for changing the report's place format to a
    report-specific format instead of the user's Edit=>Preferences choice
    """
    place_format = EnumeratedListOption(_("Place format"), -1)
    place_format.add_item(-1, _("Default"))
    for number, fmt in enumerate(_pd.get_formats()):
        place_format.add_item(number, fmt.name)
    place_format.set_help(_("Select the format to display places"))
    menu.add_option(category, "place_format", place_format)
    return place_format


def add_coordinates_format_option(menu, category):
    """
    Insert an option for changing the report's coordinates format to a
    report-specific format instead of the user's Edit=>Preferences choice
    """
    coord_format = EnumeratedListOption(_("Coordinates format"), 0)
    for number, fmt in enumerate(coord_formats):
        coord_format.add_item(number, fmt + "\t" + coord_formats_desc[number])
    coord_format.set_help(_("Select the format to display coordinates"))
    menu.add_option(category, "coord_format", coord_format)
    return coord_format



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
#
"""
Display a person's events, both personal and family
"""

from gramps.gen.simple import SimpleAccess, SimpleDoc
from gramps.gui.plug.quick import QuickTable
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


def run(database, document, person):
    """
    Loops through the person events and the family events of any family
    in which the person is a parent (to catch Marriage events), displaying
    the basic details of the event
    """

    sdb = SimpleAccess(database)
    sdoc = SimpleDoc(document)
    stab = QuickTable(sdb)

    # get the personal events
    event_list = sdb.events(person)

    # get the events of each family in which the person is
    # a parent
    for family in sdb.parent_in(person):
        event_list += sdb.events(family)

    # Sort the events by their date
    event_list.sort(key=lambda x: x.get_date_object())

    # display the results

    # feature request 2356: avoid genitive form
    sdoc.title(_("Sorted events of %s") % sdb.name(person))
    sdoc.paragraph("")

    stab.columns(_("Event Type"), _("Event Date"), _("Event Place"))
    document.has_data = False
    for event in event_list:
        stab.row(event, sdb.event_date_obj(event), sdb.event_place(event))
        document.has_data = True
    if document.has_data:
        stab.write(sdoc)
    else:
        sdoc.header1(_("Not found"))


def run_fam(database, document, family):
    """
    Loops through the family events and the events of all parents, displaying
    the basic details of the event
    """

    sdb = SimpleAccess(database)
    sdoc = SimpleDoc(document)
    stab = QuickTable(sdb)

    # get the family events
    event_list = [(family, x) for x in sdb.events(family)]

    # get the events of father and mother
    # fathername = sdb.first_name(sdb.father(family))
    event_list += [(sdb.father(family), x) for x in sdb.events(sdb.father(family))]
    # mothername = sdb.first_name(sdb.mother(family))
    event_list += [(sdb.mother(family), x) for x in sdb.events(sdb.mother(family))]

    # children events
    event_list_children = []
    for child in sdb.children(family):
        # name = sdb.first_name(child)
        event_list_children += [(child, x) for x in sdb.events(child)]

    # Sort the events by their date
    event_list.sort(key=lambda x: x[1].get_date_object())
    event_list_children.sort(key=lambda x: x[1].get_date_object())

    # display the results

    sdoc.title(
        _("Sorted events of family\n %(father)s - %(mother)s")
        % {
            "father": sdb.name(sdb.father(family)),
            "mother": sdb.name(sdb.mother(family)),
        }
    )
    sdoc.paragraph("")

    document.has_data = False
    stab.columns(_("Family Member"), _("Event Type"), _("Event Date"), _("Event Place"))

    for person, event in event_list:
        stab.row(
            person,
            sdb.event_type(event),
            sdb.event_date_obj(event),
            sdb.event_place(event),
        )
        document.has_data = True
    if document.has_data:
        stab.write(sdoc)
    else:
        sdoc.header1(_("Not found") + "\n")

    stab = QuickTable(sdb)
    sdoc.header1(_("Personal events of the children"))
    stab.columns(_("Family Member"), _("Event Type"), _("Event Date"), _("Event Place"))
    for person, event in event_list_children:
        stab.row(
            person,
            sdb.event_type(event),
            sdb.event_date_obj(event),
            sdb.event_place(event),
        )
        document.has_data = True
    if document.has_data:
        stab.write(sdoc)
    else:
        sdoc.header1(_("Not found") + "\n")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2017-2018 Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
""" LaTeX Genealogy Tree adapter for Trees """
# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod
import os
import shutil
import re
from subprocess import Popen, PIPE
from io import StringIO
import tempfile
import logging

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...utils.file import search_for
from ...lib import Person, EventType, EventRoleType, Date
from ...display.place import displayer as _pd
from ...utils.file import media_path_full
from . import BaseDoc, PAPER_PORTRAIT
from ..menu import NumberOption, TextOption, EnumeratedListOption
from ...constfunc import win
from ...config import config
from ...errors import ReportError
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
LOG = logging.getLogger(".treedoc")

# -------------------------------------------------------------------------
#
# Private Constants
#
# -------------------------------------------------------------------------
_DETAIL = [
    {"name": _("Full"), "value": "full"},
    {"name": _("Medium"), "value": "medium"},
    {"name": _("Short"), "value": "short"},
]

_NAME_FORMAT = [
    {"name": _("Given Nickname Surname"), "value": "1"},
    {"name": _("Surname Given Nickname"), "value": "2"},
    {"name": _("Surname, Given Nickname"), "value": "3"},
    {"name": _("GivenNicknameSurname"), "value": "4"},
]

_PREFERRED_NAME = [
    {"name": _("Auto"), "value": "1"},
    {"name": _("Call Name Only"), "value": "2"},
    {"name": _("Whole Given Name"), "value": "3"},
]

_MARRIAGE = [
    {"name": _("Default"), "value": ""},
    {"name": _("Above"), "value": "marriage above"},
    {"name": _("Below"), "value": "marriage below"},
    {"name": _("Not shown"), "value": "no marriage"},
]

_COLOR = [
    {"name": _("None"), "value": "none"},
    {"name": _("Default"), "value": "default"},
    {"name": _("Preferences"), "value": "preferences"},
]

_TIMEFLOW = [
    {"name": _("Down (↓)"), "value": ""},
    {"name": _("Up (↑)"), "value": "up"},
    {"name": _("Right (→)"), "value": "right"},
    {"name": _("Left (←)"), "value": "left"},
]

_EDGES = [
    {"name": _("Perpendicular"), "value": ""},
    {
        "name": _("Rounded"),
        "value": "rounded",
    },
    {
        "name": _("Swing"),
        "value": "swing",
    },
    {"name": _("Mesh"), "value": "mesh"},
]

_NOTELOC = [{"name": _("Top"), "value": "t"}, {"name": _("Bottom"), "value": "b"}]

_NOTESIZE = [
    {"name": _("Tiny"), "value": "tiny"},
    {"name": _("Script"), "value": "scriptsize"},
    {"name": _("Footnote"), "value": "footnotesize"},
    {"name": _("Small"), "value": "small"},
    {"name": _("Normal"), "value": "normalsize"},
    {"name": _("Large"), "value": "large"},
    {"name": _("Very large"), "value": "Large"},
    {"name": _("Extra large"), "value": "LARGE"},
    {"name": _("Huge"), "value": "huge"},
    {"name": _("Extra huge"), "value": "Huge"},
]

if win():
    _LATEX_FOUND = search_for("lualatex.exe")
    CREATE_NO_WINDOW = 0x8000000
else:
    _LATEX_FOUND = search_for("lualatex")


def escape(text):
    lookup = {
        "&": "\\&",
        "%": "\\%",
        "$": "\\$",
        "#": "\\#",
        "_": "\\_",
        "{": "\\{",
        "}": "\\}",
        "~": "\\~{}",
        "^": "\\^{}",
        "\\": "\\textbackslash{}",
    }
    pattern = re.compile("|".join([re.escape(key) for key in lookup.keys()]))
    return pattern.sub(lambda match: lookup[match.group(0)], text)


# ------------------------------------------------------------------------------
#
# TreeOptions
#
# ------------------------------------------------------------------------------
class TreeOptions:
    """
    Defines all of the controls necessary
    to configure the genealogy tree reports.
    """

    def __init__(self):
        # Node Options
        self.detail = None
        self.marriage = None
        self.nodesize = None
        self.levelsize = None
        self.nodecolor = None

        # Tree Options
        self.timeflow = None
        self.edges = None
        self.leveldist = None

        # Note
        self.note = None
        self.noteloc = None
        self.notesize = None

    def add_menu_options(self, menu):
        """
        Add all graph related options to the menu.

        :param menu: The menu the options should be added to.
        :type menu: :class:`.Menu`
        :return: nothing
        """
        ################################
        category = _("Node Options")
        ################################

        detail = EnumeratedListOption(_("Node detail"), "full")
        for item in _DETAIL:
            detail.add_item(item["value"], item["name"])
        detail.set_help(_("Detail of information to be shown in a node."))
        menu.add_option(category, "detail", detail)
        self.detail = detail

        name_format = EnumeratedListOption(_("Name Format"), "1")
        for item in _NAME_FORMAT:
            name_format.add_item(item["value"], item["name"])
        name_format.set_help(_("Select the format to display names"))
        menu.add_option(category, "name_format", name_format)

        preferred_name = EnumeratedListOption(_("Preferred Name"), "1")
        for item in _PREFERRED_NAME:
            preferred_name.add_item(item["value"], item["name"])
        preferred_name.set_help(
            _(
                "Select the portion of the given name to be "
                "marked as preferred. Auto sets the call "
                "name if one exists, otherwise the whole "
                "given name."
            )
        )
        menu.add_option(category, "preferred_name", preferred_name)

        marriage = EnumeratedListOption(_("Marriage"), "")
        for item in _MARRIAGE:
            marriage.add_item(item["value"], item["name"])
        marriage.set_help(_("Position of marriage information."))
        menu.add_option(category, "marriage", marriage)
        self.marriage = marriage

        nodesize = NumberOption(_("Node size"), 25, 5, 100, 5)
        nodesize.set_help(
            _(
                "One dimension of a node, in mm. If the timeflow "
                "is up or down then this is the width, otherwise "
                "it is the height."
            )
        )
        menu.add_option(category, "nodesize", nodesize)
        self.nodesize = nodesize

        levelsize = NumberOption(_("Level size"), 35, 5, 100, 5)
        levelsize.set_help(
            _(
                "One dimension of a node, in mm. If the timeflow "
                "is up or down then this is the height, otherwise "
                "it is the width."
            )
        )
        menu.add_option(category, "levelsize", levelsize)
        self.levelsize = levelsize

        nodecolor = EnumeratedListOption(_("Color"), "none")
        for item in _COLOR:
            nodecolor.add_item(item["value"], item["name"])
        nodecolor.set_help(_("Node color."))
        menu.add_option(category, "nodecolor", nodecolor)
        self.nodecolor = nodecolor

        ################################
        category = _("Tree Options")
        ################################

        timeflow = EnumeratedListOption(_("Timeflow"), "")
        for item in _TIMEFLOW:
            timeflow.add_item(item["value"], item["name"])
        timeflow.set_help(_("Direction that the graph will grow over time."))
        menu.add_option(category, "timeflow", timeflow)
        self.timeflow = timeflow

        edges = EnumeratedListOption(_("Edge style"), "")
        for item in _EDGES:
            edges.add_item(item["value"], item["name"])
        edges.set_help(_("Style of the edges between nodes."))
        menu.add_option(category, "edges", edges)
        self.edges = edges

        leveldist = NumberOption(_("Level distance"), 5, 1, 20, 1)
        leveldist.set_help(
            _(
                "The minimum amount of free space, in mm, "
                "between levels.  For vertical graphs, this "
                "corresponds to spacing between rows.  For "
                "horizontal graphs, this corresponds to spacing "
                "between columns."
            )
        )
        menu.add_option(category, "leveldist", leveldist)
        self.leveldist = leveldist

        ################################
        category = _("Note")
        ################################

        note = TextOption(_("Note to add to the tree"), [""])
        note.set_help(_("This text will be added to the tree."))
        note.set_available(False)
        menu.add_option(category, "note", note)
        self.note = note

        noteloc = EnumeratedListOption(_("Note location"), "t")
        for item in _NOTELOC:
            noteloc.add_item(item["value"], item["name"])
        noteloc.set_help(_("Whether note will appear on top " "or bottom of the page."))
        menu.add_option(category, "noteloc", noteloc)
        self.noteloc = noteloc

        notesize = EnumeratedListOption(_("Note size"), "normalsize")
        for item in _NOTESIZE:
            notesize.add_item(item["value"], item["name"])
        notesize.set_help(_("The size of note text."))
        menu.add_option(category, "notesize", notesize)
        self.notesize = notesize


# ------------------------------------------------------------------------------
#
# TreeDoc
#
# ------------------------------------------------------------------------------
class TreeDoc(metaclass=ABCMeta):
    """
    Abstract Interface for genealogy tree document generators. Output formats
    for genealogy tree reports must implement this interface to be used by the
    report system.
    """

    @abstractmethod
    def start_tree(self, option_list):
        """
        Write the start of a tree.
        """

    @abstractmethod
    def end_tree(self):
        """
        Write the end of a tree.
        """

    @abstractmethod
    def start_subgraph(self, level, subgraph_type, family, option_list=None):
        """
        Write the start of a subgraph.
        """

    @abstractmethod
    def end_subgraph(self, level):
        """
        Write the end of a subgraph.
        """

    @abstractmethod
    def write_node(self, db, level, node_type, person, marriage_flag, option_list=None):
        """
        Write the contents of a node.
        """


# ------------------------------------------------------------------------------
#
# TreeDocBase
#
# ------------------------------------------------------------------------------
class TreeDocBase(BaseDoc, TreeDoc):
    """
    Base document generator for all Graphviz document generators. Classes that
    inherit from this class will only need to implement the close function.
    The close function will generate the actual file of the appropriate type.
    """

    def __init__(self, options, paper_style):
        BaseDoc.__init__(self, None, paper_style)

        self._filename = None
        self._tex = StringIO()
        self._paper = paper_style

        get_option = options.menu.get_option_by_name

        self.detail = get_option("detail").get_value()
        self.name_format = get_option("name_format").get_value()
        self.preferred_name = get_option("preferred_name").get_value()
        self.marriage = get_option("marriage").get_value()
        self.nodesize = get_option("nodesize").get_value()
        self.levelsize = get_option("levelsize").get_value()
        self.nodecolor = get_option("nodecolor").get_value()

        self.timeflow = get_option("timeflow").get_value()
        self.edges = get_option("edges").get_value()
        self.leveldist = get_option("leveldist").get_value()

        self.note = get_option("note").get_value()
        self.noteloc = get_option("noteloc").get_value()
        self.notesize = get_option("notesize").get_value()

    def write_start(self):
        """
        Write the start of the document.
        """
        paper_size = self._paper.get_size()
        name = paper_size.get_name().lower()
        if name == "custom size":
            width = str(paper_size.get_width())
            height = str(paper_size.get_height())
            paper = "papersize={%scm,%scm}" % (width, height)
        elif name in ("a", "b", "c", "d", "e"):
            paper = "ansi" + name + "paper"
        else:
            paper = name + "paper"

        if self._paper.get_orientation() == PAPER_PORTRAIT:
            orientation = "portrait"
        else:
            orientation = "landscape"

        lmargin = self._paper.get_left_margin()
        rmargin = self._paper.get_right_margin()
        tmargin = self._paper.get_top_margin()
        bmargin = self._paper.get_bottom_margin()
        if lmargin == rmargin == tmargin == bmargin:
            margin = "margin=%scm" % lmargin
        else:
            if lmargin == rmargin:
                margin = "hmargin=%scm" % lmargin
            else:
                margin = "hmargin={%scm,%scm}" % (lmargin, rmargin)
            if tmargin == bmargin:
                margin += ",vmargin=%scm" % tmargin
            else:
                margin += ",vmargin={%scm,%scm}" % (tmargin, bmargin)

        self.write(0, "\\documentclass[%s]{article}\n" % orientation)

        self.write(0, "\\IfFileExists{libertine.sty}{\n")
        self.write(0, "    \\usepackage{libertine}\n")
        self.write(0, "}{}\n")

        self.write(0, "\\usepackage[%s,%s]{geometry}\n" % (paper, margin))
        self.write(0, "\\usepackage[all]{genealogytree}\n")
        self.write(0, "\\usepackage{color}\n")
        self.write(0, "\\begin{document}\n")

        if self.nodecolor == "preferences":
            scheme = config.get("colors.scheme")
            male_bg = config.get("colors.male-dead")[scheme][1:]
            female_bg = config.get("colors.female-dead")[scheme][1:]
            neuter_bg = config.get("colors.unknown-dead")[scheme][1:]
            self.write(0, "\\definecolor{male-bg}{HTML}{%s}\n" % male_bg)
            self.write(0, "\\definecolor{female-bg}{HTML}{%s}\n" % female_bg)
            self.write(0, "\\definecolor{neuter-bg}{HTML}{%s}\n" % neuter_bg)

        if "".join(self.note) != "" and self.noteloc == "t":
            for line in self.note:
                self.write(0, "{\\%s %s}\\par\n" % (self.notesize, line))
            self.write(0, "\\bigskip\n")

        self.write(0, "\\begin{tikzpicture}\n")

    def start_tree(self, option_list):
        self.write(0, "\\genealogytree[\n")
        self.write(0, "processing=database,\n")
        if self.marriage:
            info = self.detail + " " + self.marriage
        else:
            info = self.detail
        self.write(0, "database format=%s,\n" % info)
        if self.timeflow:
            self.write(0, "timeflow=%s,\n" % self.timeflow)
        if self.edges:
            self.write(0, "edges=%s,\n" % self.edges)
        if self.leveldist != 5:
            self.write(0, "level distance=%smm,\n" % self.leveldist)
        if self.nodesize != 25:
            self.write(0, "node size=%smm,\n" % self.nodesize)
        if self.levelsize != 35:
            self.write(0, "level size=%smm,\n" % self.levelsize)
        if self.nodecolor == "none":
            self.write(0, "tcbset={male/.style={},\n")
            self.write(0, "        female/.style={},\n")
            self.write(0, "        neuter/.style={}},\n")
        if self.nodecolor == "preferences":
            self.write(0, "tcbset={male/.style={colback=male-bg},\n")
            self.write(0, "        female/.style={colback=female-bg},\n")
            self.write(0, "        neuter/.style={colback=neuter-bg}},\n")

        for option in option_list:
            self.write(0, "%s,\n" % option)

        self.write(0, "]{\n")

    def end_tree(self):
        self.write(0, "}\n")

    def write_end(self):
        """
        Write the end of the document.
        """
        self.write(0, "\\end{tikzpicture}\n")

        if "".join(self.note) != "" and self.noteloc == "b":
            self.write(0, "\\bigskip\n")
            for line in self.note:
                self.write(0, "\\par{\\%s %s}\n" % (self.notesize, line))

        self.write(0, "\\end{document}\n")

    def start_subgraph(self, level, subgraph_type, family, option_list=None):
        options = ["id=%s" % family.gramps_id]
        if option_list:
            options.extend(option_list)
        self.write(level, "%s[%s]{\n" % (subgraph_type, ",".join(options)))

    def end_subgraph(self, level):
        self.write(level, "}\n")

    def write_node(self, db, level, node_type, person, marriage_flag, option_list=None):
        options = ["id=%s" % person.gramps_id]
        if option_list:
            options.extend(option_list)
        self.write(level, "%s[%s]{\n" % (node_type, ",".join(options)))
        if person.gender == Person.MALE:
            self.write(level + 1, "male,\n")
        elif person.gender == Person.FEMALE:
            self.write(level + 1, "female,\n")
        elif person.gender in (Person.UNKNOWN, Person.OTHER):
            self.write(level + 1, "neuter,\n")
        name = person.get_primary_name()
        nick = name.get_nick_name()
        surn = name.get_surname()
        if self.name_format == "1":
            name_parts = [
                self.format_given_names(name),
                "\\nick{{{}}}".format(escape(nick)) if nick else "",
                "\\surn{{{}}}".format(escape(surn)) if surn else "",
            ]
            self.write(
                level + 1,
                "name = {{{}}},\n".format(" ".join([e for e in name_parts if e])),
            )
        elif self.name_format == "2":
            name_parts = [
                "\\surn{{{}}}".format(escape(surn)) if surn else "",
                self.format_given_names(name),
                "\\nick{{{}}}".format(escape(nick)) if nick else "",
            ]
            self.write(
                level + 1,
                "name = {{{}}},\n".format(" ".join([e for e in name_parts if e])),
            )
        elif self.name_format == "3":
            name_parts1 = [
                self.format_given_names(name),
                "\\nick{{{}}}".format(escape(nick)) if nick else "",
            ]
            given_nick = " ".join([e for e in name_parts1 if e])
            name_parts2 = [
                "\\surn{{{}}}".format(escape(surn)) if surn else "",
                given_nick,
            ]
            self.write(
                level + 1,
                "name = {{{}}},\n".format(", ".join([e for e in name_parts2 if e])),
            )
        elif self.name_format == "4":
            name_parts = [
                self.format_given_names(name),
                "\\nick{{{}}}".format(escape(nick)) if nick else "",
                "\\surn{{{}}}".format(escape(surn)) if surn else "",
            ]
            self.write(
                level + 1,
                "name = {{{}}},\n".format("".join([e for e in name_parts if e])),
            )
        for eventref in person.get_event_ref_list():
            if eventref.role == EventRoleType.PRIMARY:
                event = db.get_event_from_handle(eventref.ref)
                self.write_event(db, level + 1, event)
        if marriage_flag:
            for handle in person.get_family_handle_list():
                family = db.get_family_from_handle(handle)
                for eventref in family.get_event_ref_list():
                    if eventref.role == EventRoleType.FAMILY:
                        event = db.get_event_from_handle(eventref.ref)
                        self.write_event(db, level + 1, event)
        for attr in person.get_attribute_list():
            # Comparison with 'Occupation' for backwards compatibility with Gramps 5.0
            attr_type = str(attr.get_type())
            if attr_type in ("Occupation", _("Occupation")):
                self.write(level + 1, "profession = {%s},\n" % escape(attr.get_value()))
            if attr_type == "Comment":
                self.write(level + 1, "comment = {%s},\n" % escape(attr.get_value()))
        for mediaref in person.get_media_list():
            media = db.get_media_from_handle(mediaref.ref)
            path = media_path_full(db, media.get_path())
            if os.path.isfile(path):
                if win():
                    path = path.replace("\\", "/")
                self.write(level + 1, "image = {%s},\n" % path)
                break  # first image only
        self.write(level, "}\n")

    def write_event(self, db, level, event):
        """
        Write an event.
        """
        modifier = None
        if event.type == EventType.BIRTH:
            event_type = "birth"
            if "died" in event.description.lower():
                modifier = "died"
            if "stillborn" in event.description.lower():
                modifier = "stillborn"
            # modifier = 'out of wedlock'
        elif event.type == EventType.BAPTISM:
            event_type = "baptism"
        elif event.type == EventType.ENGAGEMENT:
            event_type = "engagement"
        elif event.type == EventType.MARRIAGE:
            event_type = "marriage"
        elif event.type == EventType.DIVORCE:
            event_type = "divorce"
        elif event.type == EventType.DEATH:
            event_type = "death"
        elif event.type == EventType.BURIAL:
            event_type = "burial"
            if "killed" in event.description.lower():
                modifier = "killed"
        elif event.type == EventType.CREMATION:
            event_type = "burial"
            modifier = "cremated"
        else:
            return

        date = event.get_date_object()

        if date.get_calendar() == Date.CAL_GREGORIAN:
            calendar = "AD"  # GR
        elif date.get_calendar() == Date.CAL_JULIAN:
            calendar = "JU"
        else:
            calendar = ""

        if date.get_modifier() == Date.MOD_ABOUT:
            calendar = "ca" + calendar

        date_str = self.format_iso(date.get_ymd(), calendar)
        if date.get_modifier() == Date.MOD_BEFORE:
            date_str = "/" + date_str
        elif date.get_modifier() == Date.MOD_AFTER:
            date_str = date_str + "/"
        elif date.is_compound():
            stop_date = self.format_iso(date.get_stop_ymd(), calendar)
            date_str = date_str + "/" + stop_date

        place = escape(_pd.display_event(db, event))
        place = place.replace("-", "\\--")

        if modifier:
            event_type += "+"
            self.write(
                level, "%s = {%s}{%s}{%s},\n" % (event_type, date_str, place, modifier)
            )
        elif place == "":
            event_type += "-"
            self.write(level, "%s = {%s},\n" % (event_type, date_str))
        else:
            self.write(level, "%s = {%s}{%s},\n" % (event_type, date_str, place))

    def format_given_names(self, name):
        """
        Format given names.
        """
        first = name.get_first_name()
        call = name.get_call_name()
        if call and self.preferred_name != "3":
            if call in first:
                where = first.index(call)
                return "{before}\\pref{{{call}}}{after}".format(
                    before=escape(first[:where]),
                    call=escape(call),
                    after=escape(first[where + len(call) :]),
                )
        elif self.preferred_name in ("1", "3"):
            return "\\pref{{{}}}".format(escape(first))
        elif self.preferred_name == "2":
            return escape(first)

    def format_iso(self, date_tuple, calendar):
        """
        Format an iso date.
        """
        year, month, day = date_tuple
        if year == 0:
            iso_date = ""
        elif month == 0:
            iso_date = str(year)
        elif day == 0:
            iso_date = "%s-%s" % (year, month)
        else:
            iso_date = "%s-%s-%s" % (year, month, day)
        if calendar and calendar != "AD":
            iso_date = "(%s)%s" % (calendar, iso_date)
        return iso_date

    def write(self, level, text):
        """
        Write indented text.
        """
        self._tex.write("  " * level + text)

    def open(self, filename):
        """Implement TreeDocBase.open()"""
        self._filename = os.path.normpath(os.path.abspath(filename))
        self.write_start()

    def close(self):
        """
        This isn't useful by itself. Other classes need to override this and
        actually generate a file.
        """
        self.write_end()


# ------------------------------------------------------------------------------
#
# TreeGraphDoc
#
# ------------------------------------------------------------------------------
class TreeGraphDoc(TreeDocBase):
    """
    TreeGraphDoc implementation that generates a .graph file.
    """

    def write_start(self):
        """
        Write the start of the document - nothing for a graph file.
        """
        pass

    def start_tree(self, option_list):
        """
        Write the start of a tree - nothing for a graph file.
        """
        pass

    def end_tree(self):
        """
        Write the end of a tree - nothing for a graph file.
        """
        pass

    def write_end(self):
        """
        Write the end of the document - nothing for a graph file.
        """
        pass

    def close(self):
        """Implements TreeDocBase.close()"""
        TreeDocBase.close(self)

        with open(self._filename, "w", encoding="utf-8") as texfile:
            texfile.write(self._tex.getvalue())


# ------------------------------------------------------------------------------
#
# TreeTexDoc
#
# ------------------------------------------------------------------------------
class TreeTexDoc(TreeDocBase):
    """
    TreeTexDoc implementation that generates a .tex file.
    """

    def close(self):
        """Implements TreeDocBase.close()"""
        TreeDocBase.close(self)

        # Make sure the extension is correct
        if self._filename[-4:] != ".tex":
            self._filename += ".tex"

        with open(self._filename, "w", encoding="utf-8") as texfile:
            texfile.write(self._tex.getvalue())


# ------------------------------------------------------------------------------
#
# TreePdfDoc
#
# ------------------------------------------------------------------------------
class TreePdfDoc(TreeDocBase):
    """
    TreePdfDoc implementation that generates a .pdf file.
    """

    def close(self):
        """Implements TreeDocBase.close()"""
        TreeDocBase.close(self)

        # Make sure the extension is correct
        if self._filename[-4:] != ".pdf":
            self._filename += ".pdf"

        with tempfile.TemporaryDirectory() as tmpdir:
            basename = os.path.basename(self._filename)
            args = ["lualatex", "--jobname", basename[:-4]]
            if win():
                proc = Popen(
                    args,
                    stdin=PIPE,
                    stdout=PIPE,
                    stderr=PIPE,
                    cwd=tmpdir,
                    creationflags=CREATE_NO_WINDOW,
                )
            else:
                proc = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE, cwd=tmpdir)
            proc.communicate(input=self._tex.getvalue().encode("utf-8"))

            temp_output_file = os.path.join(tmpdir, basename)
            if os.path.isfile(temp_output_file):
                shutil.copy(temp_output_file, self._filename)
            else:
                raise ReportError(
                    _("Empty report"), _("Could not create %s") % self._filename
                )


# ------------------------------------------------------------------------------
#
# Various Genealogy Tree formats.
#
# ------------------------------------------------------------------------------
FORMATS = []

if _LATEX_FOUND:
    FORMATS += [
        {
            "type": "pdf",
            "ext": "pdf",
            "descr": _("PDF"),
            "mime": "application/pdf",
            "class": TreePdfDoc,
        }
    ]

FORMATS += [
    {
        "type": "graph",
        "ext": "graph",
        "descr": _("Graph File for genealogytree"),
        "class": TreeGraphDoc,
    }
]

FORMATS += [
    {
        "type": "tex",
        "ext": "tex",
        "descr": _("LaTeX File"),
        "mime": "application/x-latex",
        "class": TreeTexDoc,
    }
]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2013       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide the different event types
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .grampstype import GrampsType
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


class EventType(GrampsType):
    """
    Event types.

    .. attribute UNKNOWN:    Unknown
    .. attribute CUSTOM:    Custom
    .. attribute ADOPT:     Adopted
    .. attribute BIRTH:     Birth
    .. attribute DEATH:     Death
    .. attribute ADULT_CHRISTEN: Adult Christening
    .. attribute BAPTISM:      Baptism
    .. attribute BAR_MITZVAH:  Bar Mitzvah
    .. attribute BAS_MITZVAH:  Bas Mitzvah
    .. attribute BLESS:        Blessing
    .. attribute BURIAL:       Burial
    .. attribute CAUSE_DEATH:   Cause Of Death
    .. attribute CENSUS:        Census
    .. attribute CHRISTEN:      Christening
    .. attribute CONFIRMATION:   Confirmation
    .. attribute CREMATION:      Cremation
    .. attribute DEGREE:         Degree
    .. attribute EDUCATION:     Education
    .. attribute ELECTED:       Elected
    .. attribute EMIGRATION:    Emigration
    .. attribute FIRST_COMMUN:  First Communion
    .. attribute IMMIGRATION:   Immigration
    .. attribute GRADUATION:   Graduation
    .. attribute MED_INFO:     Medical Information
    .. attribute MILITARY_SERV:   Military Service
    .. attribute NATURALIZATION: Naturalization
    .. attribute NOB_TITLE:     Nobility Title
    .. attribute NUM_MARRIAGES:   Number of Marriages
    .. attribute OCCUPATION:      Occupation
    .. attribute ORDINATION:     Ordination
    .. attribute PROBATE:        Probate
    .. attribute PROPERTY:      Property
    .. attribute RELIGION:      Religion
    .. attribute RESIDENCE:    Residence
    .. attribute RETIREMENT:    Retirement
    .. attribute WILL:           Will
    .. attribute MARRIAGE:     Marriage
    .. attribute MARR_SETTL:     Marriage Settlement
    .. attribute MARR_LIC:       Marriage License
    .. attribute MARR_CONTR:     Marriage Contract
    .. attribute MARR_BANNS:     Marriage Banns
    .. attribute ENGAGEMENT:     Engagement
    .. attribute DIVORCE:        Divorce
    .. attribute DIV_FILING:     Divorce Filing
    .. attribute ANNULMENT:      Annulment
    .. attribute MARR_ALT:        Alternate Marriage
    .. attribute STILLBIRTH:      Stillbirth
    """

    UNKNOWN = -1
    CUSTOM = 0
    MARRIAGE = 1
    MARR_SETTL = 2
    MARR_LIC = 3
    MARR_CONTR = 4
    MARR_BANNS = 5
    ENGAGEMENT = 6
    DIVORCE = 7
    DIV_FILING = 8
    ANNULMENT = 9
    MARR_ALT = 10
    ADOPT = 11
    BIRTH = 12
    DEATH = 13
    ADULT_CHRISTEN = 14
    BAPTISM = 15
    BAR_MITZVAH = 16
    BAS_MITZVAH = 17
    BLESS = 18
    BURIAL = 19
    CAUSE_DEATH = 20
    CENSUS = 21
    CHRISTEN = 22
    CONFIRMATION = 23
    CREMATION = 24
    DEGREE = 25
    EDUCATION = 26
    ELECTED = 27
    EMIGRATION = 28
    FIRST_COMMUN = 29
    IMMIGRATION = 30
    GRADUATION = 31
    MED_INFO = 32
    MILITARY_SERV = 33
    NATURALIZATION = 34
    NOB_TITLE = 35
    NUM_MARRIAGES = 36
    OCCUPATION = 37
    ORDINATION = 38
    PROBATE = 39
    PROPERTY = 40
    RELIGION = 41
    RESIDENCE = 42
    RETIREMENT = 43
    WILL = 44
    STILLBIRTH = 45

    # _T_ is a gramps-defined keyword -- see po/update_po.py and po/genpot.sh
    def _T_(value, context=""):  # enable deferred translations
        return "%s\x04%s" % (context, value) if context else value

    _MENU = [
        [
            _T_("Life Events"),
            [BIRTH, BAPTISM, DEATH, STILLBIRTH, BURIAL, CREMATION, ADOPT],
        ],
        [
            _T_("Family"),
            [
                ENGAGEMENT,
                MARRIAGE,
                DIVORCE,
                ANNULMENT,
                MARR_SETTL,
                MARR_LIC,
                MARR_CONTR,
                MARR_BANNS,
                DIV_FILING,
                MARR_ALT,
            ],
        ],
        [
            _T_("Religious"),
            [
                CHRISTEN,
                ADULT_CHRISTEN,
                CONFIRMATION,
                FIRST_COMMUN,
                BLESS,
                BAR_MITZVAH,
                BAS_MITZVAH,
                RELIGION,
            ],
        ],
        [
            _T_("Vocational"),
            [OCCUPATION, RETIREMENT, ELECTED, MILITARY_SERV, ORDINATION],
        ],
        [_T_("Academic"), [EDUCATION, DEGREE, GRADUATION]],
        [_T_("Travel"), [EMIGRATION, IMMIGRATION, NATURALIZATION]],
        [_T_("Legal"), [PROBATE, WILL]],
        [_T_("Residence"), [RESIDENCE, CENSUS, PROPERTY]],
        [_T_("Other"), [CAUSE_DEATH, MED_INFO, NOB_TITLE, NUM_MARRIAGES]],
    ]

    _CUSTOM = CUSTOM
    _DEFAULT = BIRTH

    _DATAMAP = [
        (UNKNOWN, _("Unknown"), "Unknown"),
        (CUSTOM, _("Custom"), "Custom"),
        (ADOPT, _("Adopted"), "Adopted"),
        (BIRTH, _("Birth"), "Birth"),
        (DEATH, _("Death"), "Death"),
        (ADULT_CHRISTEN, _("Adult Christening"), "Adult Christening"),
        (BAPTISM, _("Baptism"), "Baptism"),
        (BAR_MITZVAH, _("Bar Mitzvah"), "Bar Mitzvah"),
        (BAS_MITZVAH, _("Bat Mitzvah"), "Bas Mitzvah"),
        (BLESS, _("Blessing"), "Blessing"),
        (BURIAL, _("Burial"), "Burial"),
        (CAUSE_DEATH, _("Cause Of Death"), "Cause Of Death"),
        (CENSUS, _("Census"), "Census"),
        (CHRISTEN, _("Christening"), "Christening"),
        (CONFIRMATION, _("Confirmation"), "Confirmation"),
        (CREMATION, _("Cremation"), "Cremation"),
        (DEGREE, _("Degree"), "Degree"),
        (EDUCATION, _("Education"), "Education"),
        (ELECTED, _("Elected"), "Elected"),
        (EMIGRATION, _("Emigration"), "Emigration"),
        (FIRST_COMMUN, _("First Communion"), "First Communion"),
        (IMMIGRATION, _("Immigration"), "Immigration"),
        (GRADUATION, _("Graduation"), "Graduation"),
        (MED_INFO, _("Medical Information"), "Medical Information"),
        (MILITARY_SERV, _("Military Service"), "Military Service"),
        (NATURALIZATION, _("Naturalization"), "Naturalization"),
        (NOB_TITLE, _("Nobility Title"), "Nobility Title"),
        (NUM_MARRIAGES, _("Number of Marriages"), "Number of Marriages"),
        (OCCUPATION, _("Occupation"), "Occupation"),
        (ORDINATION, _("Ordination"), "Ordination"),
        (PROBATE, _("Probate"), "Probate"),
        (PROPERTY, _("Property"), "Property"),
        (RELIGION, _("Religion"), "Religion"),
        (RESIDENCE, _("Residence"), "Residence"),
        (RETIREMENT, _("Retirement"), "Retirement"),
        (WILL, _("Will"), "Will"),
        (MARRIAGE, _("Marriage"), "Marriage"),
        (MARR_SETTL, _("Marriage Settlement"), "Marriage Settlement"),
        (MARR_LIC, _("Marriage License"), "Marriage License"),
        (MARR_CONTR, _("Marriage Contract"), "Marriage Contract"),
        (MARR_BANNS, _("Marriage Banns"), "Marriage Banns"),
        (ENGAGEMENT, _("Engagement"), "Engagement"),
        (DIVORCE, _("Divorce"), "Divorce"),
        (DIV_FILING, _("Divorce Filing"), "Divorce Filing"),
        (ANNULMENT, _("Annulment"), "Annulment"),
        (MARR_ALT, _("Alternate Marriage"), "Alternate Marriage"),
        (STILLBIRTH, _("Stillbirth"), "Stillbirth"),
    ]

    _ABBREVIATIONS = {
        BIRTH: _T_("b.", "birth abbreviation"),
        DEATH: _T_("d.", "death abbreviation"),
        MARRIAGE: _T_("m.", "marriage abbreviation"),
        UNKNOWN: _T_("unkn.", "Unknown abbreviation"),
        CUSTOM: _T_("cust.", "Custom abbreviation"),
        ADOPT: _T_("adop.", "Adopted abbreviation"),
        ADULT_CHRISTEN: _T_("a.chr.", "Adult Christening abbreviation"),
        BAPTISM: _T_("bap.", "Baptism abbreviation"),
        BAR_MITZVAH: _T_("bar.", "Bar Mitzvah abbreviation"),
        BAS_MITZVAH: _T_("bat.", "Bat Mitzvah abbreviation"),
        BLESS: _T_("bles.", "Blessing abbreviation"),
        BURIAL: _T_("bur.", "Burial abbreviation"),
        CAUSE_DEATH: _T_("d.cau.", "Cause Of Death abbreviation"),
        CENSUS: _T_("cens.", "Census abbreviation"),
        CHRISTEN: _T_("chr.", "Christening abbreviation"),
        CONFIRMATION: _T_("conf.", "Confirmation abbreviation"),
        CREMATION: _T_("crem.", "Cremation abbreviation"),
        DEGREE: _T_("deg.", "Degree abbreviation"),
        EDUCATION: _T_("edu.", "Education abbreviation"),
        ELECTED: _T_("elec.", "Elected abbreviation"),
        EMIGRATION: _T_("em.", "Emigration abbreviation"),
        FIRST_COMMUN: _T_("f.comm.", "First Communion abbreviation"),
        IMMIGRATION: _T_("im.", "Immigration abbreviation"),
        GRADUATION: _T_("grad.", "Graduation abbreviation"),
        MED_INFO: _T_("medinf.", "Medical Information abbreviation"),
        MILITARY_SERV: _T_("milser.", "Military Service abbreviation"),
        NATURALIZATION: _T_("nat.", "Naturalization abbreviation"),
        NOB_TITLE: _T_("nob.", "Nobility Title abbreviation"),
        NUM_MARRIAGES: _T_("n.o.mar.", "Number of Marriages abbreviation"),
        OCCUPATION: _T_("occ.", "Occupation abbreviation"),
        ORDINATION: _T_("ord.", "Ordination abbreviation"),
        PROBATE: _T_("prob.", "Probate abbreviation"),
        PROPERTY: _T_("prop.", "Property abbreviation"),
        RELIGION: _T_("rel.", "Religion abbreviation"),
        RESIDENCE: _T_("res.", "Residence abbreviation"),
        RETIREMENT: _T_("ret.", "Retirement abbreviation"),
        WILL: _T_("will.", "Will abbreviation"),
        MARR_SETTL: _T_("m.set.", "Marriage Settlement abbreviation"),
        MARR_LIC: _T_("m.lic.", "Marriage License abbreviation"),
        MARR_CONTR: _T_("m.con.", "Marriage Contract abbreviation"),
        MARR_BANNS: _T_("m.ban.", "Marriage Banns abbreviation"),
        MARR_ALT: _T_("alt.mar.", "Alternate Marriage abbreviation"),
        ENGAGEMENT: _T_("engd.", "Engagement abbreviation"),
        DIVORCE: _T_("div.", "Divorce abbreviation"),
        DIV_FILING: _T_("div.f.", "Divorce Filing abbreviation"),
        ANNULMENT: _T_("annul.", "Annulment abbreviation"),
        STILLBIRTH: _T_("still.", "Stillbirth abbreviation"),
    }

    def __init__(self, value=None):
        GrampsType.__init__(self, value)

    def is_birth(self):
        """
        Returns True if EventType is BIRTH, False
        otherwise.
        """
        return self.value == self.BIRTH

    def is_baptism(self):
        """
        Returns True if EventType is BAPTISM, False
        otherwise.
        """
        return self.value == self.BAPTISM

    def is_death(self):
        """
        Returns True if EventType is DEATH, False
        otherwise.
        """
        return self.value == self.DEATH

    def is_burial(self):
        """
        Returns True if EventType is BURIAL, False
        otherwise.
        """
        return self.value == self.BURIAL

    def is_birth_fallback(self):
        """
        Returns True if EventType is a birth fallback, False
        otherwise.
        """
        return self.value in [self.STILLBIRTH, self.BAPTISM, self.CHRISTEN]

    def is_death_fallback(self):
        """
        Returns True if EventType is a death fallback, False
        otherwise.
        """
        return self.value in [
            self.STILLBIRTH,
            self.BURIAL,
            self.CREMATION,
            self.CAUSE_DEATH,
        ]

    def is_marriage(self):
        """
        Returns True if EventType is MARRIAGE, False otherwise.
        """
        return self.value == self.MARRIAGE

    def is_marriage_fallback(self):
        """
        Returns True if EventType is a marriage fallback, False
        otherwise.
        """
        return self.value in [self.ENGAGEMENT, self.MARR_ALT]

    def is_divorce(self):
        """
        Returns True if EventType is DIVORCE, False otherwise.
        """
        return self.value == self.DIVORCE

    def is_divorce_fallback(self):
        """
        Returns True if EventType is a divorce fallback, False
        otherwise.
        """
        return self.value in [self.ANNULMENT, self.DIV_FILING]

    def is_relationship_event(self):
        """
        Returns True is EventType is a relationship type event.
        """
        return self.value in [self.DIVORCE, self.MARRIAGE, self.ANNULMENT]

    def is_type(self, event_name):
        """
        Returns True if EventType has name EVENT_NAME, False otherwise.
        """
        event_type = [tup for tup in self._DATAMAP if tup[2] == event_name]
        if len(event_type) > 0:
            return self.value == event_type[0][0]  # first one, the code
        return False

    def get_abbreviation(self, trans_text=glocale.translation.sgettext):
        """
        Returns the abbreviation for this event. Uses the explicitly
        given abbreviations, or first letter of each word, or the first
        three letters. Appends a period after the abbreviation,
        but not if string is in _ABBREVIATIONS.

        If trans_text is passed in (a GrampsLocale sgettext instance)
        then the translated abbreviation will be returned instead.
        """
        if self.value in self._ABBREVIATIONS:
            return trans_text(self._ABBREVIATIONS[self.value])
        else:
            abbrev = str(self)
            if " " in abbrev:
                return ".".join([letter[0].lower() for letter in abbrev.split()]) + "."
            else:
                return abbrev[:3].lower() + "."



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
This module provides the base class for plugin registration.
It provides an object containing data about the plugin (version, filename, ...)
and a register for the data of all plugins .
"""
# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
import os
import sys
import re
import traceback

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...version import VERSION as GRAMPSVERSION, VERSION_TUPLE
from ..utils.requirements import Requirements
from ..const import IMAGE_DIR
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
import logging

LOG = logging.getLogger("._manager")

# Fix DEBUG for AIO built with cx_Freeze
DEBUG = __debug__ and not hasattr(sys, "frozen")

# -------------------------------------------------------------------------
#
# PluginData
#
# -------------------------------------------------------------------------

# Development status
UNSTABLE = 0
EXPERIMENTAL = 1
BETA = 2
STABLE = 3
STATUS = [UNSTABLE, EXPERIMENTAL, BETA, STABLE]
STATUSTEXT = {
    UNSTABLE: _("Unstable"),
    EXPERIMENTAL: _("Experimental"),
    BETA: _("Beta"),
    STABLE: _("Stable"),
}

# Intended audience
EVERYONE = 0
EXPERT = 1
DEVELOPER = 2
AUDIENCE = [EVERYONE, EXPERT, DEVELOPER]
AUDIENCETEXT = {EVERYONE: _("Everyone"), EXPERT: _("Expert"), DEVELOPER: _("Developer")}

# possible plugin types
REPORT = 0
QUICKREPORT = 1  # deprecated
QUICKVIEW = 1
TOOL = 2
IMPORT = 3
EXPORT = 4
DOCGEN = 5
GENERAL = 6
MAPSERVICE = 7
VIEW = 8
RELCALC = 9
GRAMPLET = 10
SIDEBAR = 11
DATABASE = 12
RULE = 13
THUMBNAILER = 14
CITE = 15
PTYPE = [
    REPORT,
    QUICKREPORT,
    TOOL,
    IMPORT,
    EXPORT,
    DOCGEN,
    GENERAL,
    MAPSERVICE,
    VIEW,
    RELCALC,
    GRAMPLET,
    SIDEBAR,
    DATABASE,
    RULE,
    THUMBNAILER,
    CITE,
]
PTYPE_STR = {
    REPORT: _("Report"),
    QUICKREPORT: _("Quick view"),
    TOOL: _("Tool"),
    IMPORT: _("Importer"),
    EXPORT: _("Exporter"),
    DOCGEN: _("Document creator"),
    GENERAL: _("Plugin library"),
    MAPSERVICE: _("Map service"),
    VIEW: _("View"),
    RELCALC: _("Relationship calculator"),
    GRAMPLET: _("Gramplet"),
    SIDEBAR: _("Sidebar"),
    DATABASE: _("Database"),
    RULE: _("Rule"),
    THUMBNAILER: _("Thumbnailer"),
    CITE: _("Citation formatter"),
}

# possible report categories
CATEGORY_TEXT = 0
CATEGORY_DRAW = 1
CATEGORY_CODE = 2
CATEGORY_WEB = 3
CATEGORY_BOOK = 4
CATEGORY_GRAPHVIZ = 5
CATEGORY_TREE = 6
REPORT_CAT = [
    CATEGORY_TEXT,
    CATEGORY_DRAW,
    CATEGORY_CODE,
    CATEGORY_WEB,
    CATEGORY_BOOK,
    CATEGORY_GRAPHVIZ,
    CATEGORY_TREE,
]
# possible tool categories
TOOL_DEBUG = -1
TOOL_ANAL = 0
TOOL_DBPROC = 1
TOOL_DBFIX = 2
TOOL_REVCTL = 3
TOOL_UTILS = 4
TOOL_CAT = [TOOL_DEBUG, TOOL_ANAL, TOOL_DBPROC, TOOL_DBFIX, TOOL_REVCTL, TOOL_UTILS]

# possible quickreport categories
CATEGORY_QR_MISC = -1
CATEGORY_QR_PERSON = 0
CATEGORY_QR_FAMILY = 1
CATEGORY_QR_EVENT = 2
CATEGORY_QR_SOURCE = 3
CATEGORY_QR_PLACE = 4
CATEGORY_QR_REPOSITORY = 5
CATEGORY_QR_NOTE = 6
CATEGORY_QR_DATE = 7
CATEGORY_QR_MEDIA = 8
CATEGORY_QR_CITATION = 9
CATEGORY_QR_SOURCE_OR_CITATION = 10

# Modes for generating reports
REPORT_MODE_GUI = 1  # Standalone report using GUI
REPORT_MODE_BKI = 2  # Book Item interface using GUI
REPORT_MODE_CLI = 4  # Command line interface (CLI)
REPORT_MODES = [REPORT_MODE_GUI, REPORT_MODE_BKI, REPORT_MODE_CLI]

# Modes for running tools
TOOL_MODE_GUI = 1  # Standard tool using GUI
TOOL_MODE_CLI = 2  # Command line interface (CLI)
TOOL_MODES = [TOOL_MODE_GUI, TOOL_MODE_CLI]

# possible view orders
START = 1
END = 2


# -------------------------------------------------------------------------
#
# Functions and classes
#
# -------------------------------------------------------------------------
def myint(s):
    """
    Protected version of int()
    """
    try:
        v = int(s)
    except:
        v = s
    return v


def version(sversion):
    """
    Return the tuple version of a string version.
    """
    return tuple([myint(x or "0") for x in (sversion + "..").split(".")])


def valid_plugin_version(plugin_version_string):
    """
    Checks to see if string is a valid version string for this version
    of Gramps.
    """
    if not isinstance(plugin_version_string, str):
        return False
    dots = plugin_version_string.count(".")
    if dots == 1:
        plugin_version = tuple(map(int, plugin_version_string.split(".", 1)))
        return plugin_version == VERSION_TUPLE[:2]
    elif dots == 2:
        plugin_version = tuple(map(int, plugin_version_string.split(".", 2)))
        return (
            plugin_version[:2] == VERSION_TUPLE[:2] and plugin_version <= VERSION_TUPLE
        )
    return False


class PluginData:
    """
    This is the base class for all plugin data objects.
    The workflow is:

    1. plugin manager reads all register files, and stores plugin data
       objects in a plugin register
    2. when plugin is needed, the plugin register creates the plugin, and
       the manager stores this, after which it can be executed.

    Attributes present for all plugins

    .. attribute:: id
       A unique identifier for the plugin. This is eg used to store the plugin
       settings.  MUST be in ASCII, with only "_- ().,'" special characters.
    .. attribute:: name
       A friendly name to call this plugin (normally translated)
    .. attribute:: name_accell
       A friendly name to call this plugin (normally translated), with an
       accellerator present (eg '_Descendant report', with D to be accellerator
       key
    .. attribute:: description
       A friendly description of what the plugin does
    .. attribute:: version
       The version of the plugin
    .. attribute:: status
       The development status of the plugin, UNSTABLE, EXPERIMENTAL, BETA or
       STABLE. UNSTABLE is only visible in development code, not in release
    .. attribute:: audience
       The intended audience of the plugin, EVERYONE, DEVELOPER or EXPERT.
    .. attribute:: fname
       The python file where the plugin implementation can be found
    .. attribute:: fpath
       The python path where the plugin implementation can be found
    .. attribute:: ptype
       The plugin type. One of REPORT , QUICKREPORT, TOOL, IMPORT,
        EXPORT, DOCGEN, GENERAL, MAPSERVICE, VIEW, GRAMPLET, DATABASE, RULE
    .. attribute:: authors
       List of authors of the plugin, default=[]
    .. attribute:: authors_email
       List of emails of the authors of the plugin, default=[]
    .. attribute:: maintainers
       List of maintainers of the plugin, default=[]
    .. attribute:: maintainers_email
       List of emails of the maintainers of the plugin, default=[]
    .. attribute:: supported
       Bool value indicating if the plugin is still supported, default=True
    .. attribute:: load_on_reg
       bool value, if True, the plugin is loaded on Gramps startup. Some
       plugins. Only set this value if for testing you want the plugin to be
       loaded immediately on startup. default=False
    .. attribute: icons
       New stock icons to register. A list of tuples (stock_id, icon_label),
       eg:
            [('gramps_myplugin', _('My Plugin')),
            ('gramps_myplugin_open', _('Open Plugin')]
       The icon directory must contain the directories scalable, 48x48, 22x22
       and 16x16 with the icons, eg:
            scalable/gramps_myplugin.svg
            48x48/gramps_myplugin.png
            22x22/gramps_myplugin.png
    .. attribute: icondir
       The directory to use for the icons. If icondir is not set or None, it
       reverts to the plugindirectory itself.
    .. attribute:: help_url
       The URL where documentation for the plugin can be found
    .. attribute:: requires_mod
       A list of required modules that should be importable using the python
       `import` statement.
    .. attribute:: requires_gi
       A list of (module, version) tuples that specify required modules that
       are reuired to be loaded via the GObject introspection system.
       eg: [('GExiv2', '0.10')]
    .. attribute:: requires_exe
       A list of executables required by the plugin. These are searched for in
       the paths specified in the PATH environment variable.

    Attributes for RELCALC plugins:

    .. attribute:: relcalcclass
       The class in the module that is the relationcalc class
    .. attribute:: lang_list
       List of languages this plugin handles

    Attributes for REPORT plugins:

    .. attribute:: require_active
       Bool, If the reports requries an active person to be set or not
    .. attribute:: reportclass
       The class in the module that is the report class
    .. attribute:: report_modes
       The report modes: list of REPORT_MODE_GUI ,REPORT_MODE_BKI,REPORT_MODE_CLI

    Attributes for REPORT and TOOL and QUICKREPORT and VIEW plugins

    .. attribute:: category
       Or the report category the plugin belongs to, default=CATEGORY_TEXT
       or the tool category a plugin belongs to, default=TOOL_UTILS
       or the quickreport category a plugin belongs to, default=CATEGORY_QR_PERSON
       or the view category a plugin belongs to,
           default=("Miscellaneous", _("Miscellaneous"))

    Attributes for REPORT and TOOL and DOCGEN plugins

    .. attribute:: optionclass
       The class in the module that is the option class

    Attributes for TOOL plugins

    .. attribute:: toolclass
       The class in the module that is the tool class
    .. attribute:: tool_modes
       The tool modes: list of TOOL_MODE_GUI, TOOL_MODE_CLI

    Attributes for DOCGEN plugins

    .. attribute :: docclass
       The class in the module that is the BaseDoc defined
    .. attribute :: paper
       bool, Indicates whether the plugin uses paper or not, default=True
    .. attribute :: style
       bool, Indicates whether the plugin uses styles or not, default=True

    Attribute for DOCGEN, EXPORT plugins

    .. attribute :: extension
       str, The file extension to use for output produced by the docgen/export,
       default=''

    Attributes for QUICKREPORT plugins

    .. attribute:: runfunc
       The function that executes the quick report

    Attributes for MAPSERVICE plugins

    .. attribute:: mapservice
       The class in the module that is a mapservice

    Attributes for EXPORT plugins

    .. attribute:: export_function
       Function that produces the export
    .. attribute:: export_options
       Class to set options
    .. attribute:: export_options_title
       Title for the option page

    Attributes for IMPORT plugins

    .. attribute:: import_function
       Function that starts an import

    Attributes for GRAMPLET plugins

    .. attribute:: gramplet
       The function or class that defines the gramplet.
    .. attribute:: height
       The height the gramplet should have in a column on GrampletView,
       default = 200
    .. attribute:: detached_height
       The height the gramplet should have detached, default 300
    .. attribute:: detached_width
       The width the gramplet should have detached, default 400
    .. attribute:: expand
       If the attributed should be expanded on start, default False
    .. attribute:: gramplet_title
       Title to use for the gramplet, default = _('Gramplet')
    .. attribute:: navtypes
       Navigation types that the gramplet is appropriate for, default = []

    Attributes for VIEW plugins

    .. attribute:: viewclass
       A class of type ViewCreator that holds the needed info of the
       view to be created: icon, viewclass that derives from pageview, ...
    .. attribute:: stock_category_icon
       The icon used for the view category if there is none
    .. attribute:: stock_icon
       The icon in the toolbar or sidebar used to select the view

    Attributes for SIDEBAR plugins

    .. attribute:: sidebarclass
       The class that defines the sidebar.
    .. attribute:: menu_label
       A label to use on the seltion menu.

    Attributes for VIEW and SIDEBAR plugins

    .. attribute:: order
       order can be START or END. Default is END. For END, on registering,
       the plugin is appended to the list of plugins. If START, then the
       plugin is prepended. Only set START if you want a plugin to be the
       first in the order of plugins

    Attributes for DATABASE plugins

    .. attribute:: databaseclass
       The class in the module that is the database class
    .. attribute:: reset_system
       Boolean to indicate that the system (sys.modules) should
       be reset.

    Attributes for RULE plugins

    .. attribute:: namespace
       The class (Person, Event, Media, etc.) the rule applies to.
    .. attribute:: ruleclass
       The exact class name of the rule; ex: HasSourceParameter

    Attributes for THUMBNAILER plugins

    .. attribute:: thumbnailer
       The exact class name of the thumbnailer
    """

    def __init__(self):
        # read/write attribute
        self.directory = None
        # base attributes
        self._id = None
        self._name = None
        self._name_accell = None
        self._version = None
        self._gramps_target_version = None
        self._description = None
        self._status = UNSTABLE
        self._audience = EVERYONE
        self._fname = None
        self._fpath = None
        self._ptype = None
        self._authors = []
        self._authors_email = []
        self._maintainers = []
        self._maintainers_email = []
        self._supported = True
        self._load_on_reg = False
        self._icons = []
        self._icondir = None
        self._depends_on = []
        self._requires_mod = []
        self._requires_gi = []
        self._requires_exe = []
        self._include_in_listing = True
        self._help_url = None
        # derived var
        self.mod_name = None
        # RELCALC attr
        self._relcalcclass = None
        self._lang_list = None
        # REPORT attr
        self._reportclass = None
        self._require_active = True
        self._report_modes = [REPORT_MODE_GUI]
        # REPORT and TOOL and GENERAL attr
        self._category = None
        # REPORT and TOOL attr
        self._optionclass = None
        # TOOL attr
        self._toolclass = None
        self._tool_modes = [TOOL_MODE_GUI]
        # DOCGEN attr
        self._paper = True
        self._style = True
        self._extension = ""
        # QUICKREPORT attr
        self._runfunc = None
        # MAPSERVICE attr
        self._mapservice = None
        # EXPORT attr
        self._export_function = None
        self._export_options = None
        self._export_options_title = ""
        # IMPORT attr
        self._import_function = None
        # GRAMPLET attr
        self._gramplet = None
        self._height = 200
        self._detached_height = 300
        self._detached_width = 400
        self._expand = False
        self._gramplet_title = _("Gramplet")
        self._navtypes = []
        self._orientation = None
        # VIEW attr
        self._viewclass = None
        self._stock_category_icon = None
        self._stock_icon = None
        # SIDEBAR attr
        self._sidebarclass = None
        self._menu_label = ""
        # VIEW and SIDEBAR attr
        self._order = END
        # DATABASE attr
        self._databaseclass = None
        self._reset_system = False
        # GENERAL attr
        self._data = []
        self._process = None
        # RULE attr
        self._ruleclass = None
        self._namespace = None
        # THUMBNAILER attr
        self._thumbnailer = None

    def _set_id(self, id):
        self._id = id

    def _get_id(self):
        return self._id

    def _set_name(self, name):
        self._name = name

    def _get_name(self):
        return self._name

    def _set_name_accell(self, name):
        self._name_accell = name

    def _get_name_accell(self):
        if self._name_accell is None:
            return self._name
        else:
            return self._name_accell

    def _set_description(self, description):
        self._description = description

    def _get_description(self):
        return self._description

    def _set_version(self, version):
        self._version = version

    def _get_version(self):
        return self._version

    def _set_gramps_target_version(self, version):
        self._gramps_target_version = version

    def _get_gramps_target_version(self):
        return self._gramps_target_version

    def _set_status(self, status):
        if status not in STATUS:
            raise ValueError("plugin status cannot be %s" % str(status))
        self._status = status

    def _get_status(self):
        return self._status

    def _set_audience(self, audience):
        if audience not in AUDIENCE:
            raise ValueError("plugin audience cannot be %s" % str(audience))
        self._audience = audience

    def _get_audience(self):
        return self._audience

    def _set_fname(self, fname):
        self._fname = fname

    def _get_fname(self):
        return self._fname

    def _set_fpath(self, fpath):
        self._fpath = fpath

    def _get_fpath(self):
        return self._fpath

    def _set_ptype(self, ptype):
        if ptype not in PTYPE:
            raise ValueError("Plugin type cannot be %s" % str(ptype))
        elif self._ptype is not None:
            raise ValueError("Plugin type may not be changed")
        self._ptype = ptype
        if self._ptype == REPORT:
            self._category = CATEGORY_TEXT
        elif self._ptype == TOOL:
            self._category = TOOL_UTILS
        elif self._ptype == QUICKREPORT:
            self._category = CATEGORY_QR_PERSON
        elif self._ptype == VIEW:
            self._category = ("Miscellaneous", _("Miscellaneous"))
        # if self._ptype == DOCGEN:
        #    self._load_on_reg = True

    def _get_ptype(self):
        return self._ptype

    def _set_authors(self, authors):
        if not authors or not isinstance(authors, list):
            return
        self._authors = authors

    def _get_authors(self):
        return self._authors

    def _set_authors_email(self, authors_email):
        if not authors_email or not isinstance(authors_email, list):
            return
        self._authors_email = authors_email

    def _get_authors_email(self):
        return self._authors_email

    def _set_maintainers(self, maintainers):
        if not maintainers or not isinstance(maintainers, list):
            return
        self._maintainers = maintainers

    def _get_maintainers(self):
        return self._maintainers

    def _set_maintainers_email(self, maintainers_email):
        if not maintainers_email or not isinstance(maintainers_email, list):
            return
        self._maintainers_email = maintainers_email

    def _get_maintainers_email(self):
        return self._maintainers_email

    def _set_supported(self, supported):
        if not isinstance(supported, bool):
            raise ValueError("Plugin must have supported=True or False")
        self._supported = supported

    def _get_supported(self):
        return self._supported

    def _set_load_on_reg(self, load_on_reg):
        if not isinstance(load_on_reg, bool):
            raise ValueError("Plugin must have load_on_reg=True or False")
        self._load_on_reg = load_on_reg

    def _get_load_on_reg(self):
        return self._load_on_reg

    def _get_icons(self):
        return self._icons

    def _set_icons(self, icons):
        if not isinstance(icons, list):
            raise ValueError("Plugin must have icons as a list")
        self._icons = icons

    def _get_icondir(self):
        return self._icondir

    def _set_icondir(self, icondir):
        self._icondir = icondir

    def _get_depends_on(self):
        return self._depends_on

    def _set_depends_on(self, depends):
        if not isinstance(depends, list):
            raise ValueError("Plugin must have depends_on as a list")
        self._depends_on = depends

    def _get_requires_mod(self):
        return self._requires_mod

    def _set_requires_mod(self, requires):
        if not isinstance(requires, list):
            raise ValueError("Plugin must have requires_mod as a list")
        self._requires_mod = requires

    def _get_requires_gi(self):
        return self._requires_gi

    def _set_requires_gi(self, requires):
        if not isinstance(requires, list):
            raise ValueError("Plugin must have requires_gi as a list")
        self._requires_gi = requires

    def _get_requires_exe(self):
        return self._requires_exe

    def _set_requires_exe(self, requires):
        if not isinstance(requires, list):
            raise ValueError("Plugin must have requires_exe as a list")
        self._requires_exe = requires

    def _get_include_in_listing(self):
        return self._include_in_listing

    def _set_include_in_listing(self, include):
        if not isinstance(include, bool):
            raise ValueError("Plugin must have include_in_listing as a bool")
        self._include_in_listing = include

    def _set_help_url(self, help_url):
        self._help_url = help_url

    def _get_help_url(self):
        return self._help_url

    id = property(_get_id, _set_id)
    name = property(_get_name, _set_name)
    name_accell = property(_get_name_accell, _set_name_accell)
    description = property(_get_description, _set_description)
    version = property(_get_version, _set_version)
    gramps_target_version = property(
        _get_gramps_target_version, _set_gramps_target_version
    )
    status = property(_get_status, _set_status)
    audience = property(_get_audience, _set_audience)
    fname = property(_get_fname, _set_fname)
    fpath = property(_get_fpath, _set_fpath)
    ptype = property(_get_ptype, _set_ptype)
    authors = property(_get_authors, _set_authors)
    authors_email = property(_get_authors_email, _set_authors_email)
    maintainers = property(_get_maintainers, _set_maintainers)
    maintainers_email = property(_get_maintainers_email, _set_maintainers_email)
    supported = property(_get_supported, _set_supported)
    load_on_reg = property(_get_load_on_reg, _set_load_on_reg)
    icons = property(_get_icons, _set_icons)
    icondir = property(_get_icondir, _set_icondir)
    depends_on = property(_get_depends_on, _set_depends_on)
    requires_mod = property(_get_requires_mod, _set_requires_mod)
    requires_gi = property(_get_requires_gi, _set_requires_gi)
    requires_exe = property(_get_requires_exe, _set_requires_exe)
    include_in_listing = property(_get_include_in_listing, _set_include_in_listing)
    help_url = property(_get_help_url, _set_help_url)

    def statustext(self):
        return STATUSTEXT[self.status]

    # type specific plugin attributes

    # RELCALC attributes
    def _set_relcalcclass(self, relcalcclass):
        if not self._ptype == RELCALC:
            raise ValueError("relcalcclass may only be set for RELCALC plugins")
        self._relcalcclass = relcalcclass

    def _get_relcalcclass(self):
        return self._relcalcclass

    def _set_lang_list(self, lang_list):
        if not self._ptype == RELCALC:
            raise ValueError("relcalcclass may only be set for RELCALC plugins")
        self._lang_list = lang_list

    def _get_lang_list(self):
        return self._lang_list

    relcalcclass = property(_get_relcalcclass, _set_relcalcclass)
    lang_list = property(_get_lang_list, _set_lang_list)

    # REPORT attributes
    def _set_require_active(self, require_active):
        if not self._ptype == REPORT:
            raise ValueError("require_active may only be set for REPORT plugins")
        if not isinstance(require_active, bool):
            raise ValueError("Report must have require_active=True or False")
        self._require_active = require_active

    def _get_require_active(self):
        return self._require_active

    def _set_reportclass(self, reportclass):
        if not self._ptype == REPORT:
            raise ValueError("reportclass may only be set for REPORT plugins")
        self._reportclass = reportclass

    def _get_reportclass(self):
        return self._reportclass

    def _set_report_modes(self, report_modes):
        if not self._ptype == REPORT:
            raise ValueError("report_modes may only be set for REPORT plugins")
        if not isinstance(report_modes, list):
            raise ValueError("report_modes must be a list")
        self._report_modes = [x for x in report_modes if x in REPORT_MODES]
        if not self._report_modes:
            raise ValueError("report_modes not a valid list of modes")

    def _get_report_modes(self):
        return self._report_modes

    # REPORT or TOOL or QUICKREPORT or GENERAL attributes
    def _set_category(self, category):
        if self._ptype not in [REPORT, TOOL, QUICKREPORT, VIEW, GENERAL]:
            raise ValueError(
                "category may only be set for "
                "REPORT/TOOL/QUICKREPORT/VIEW/GENERAL plugins"
            )
        self._category = category

    def _get_category(self):
        return self._category

    # REPORT OR TOOL attributes
    def _set_optionclass(self, optionclass):
        if not (self._ptype == REPORT or self.ptype == TOOL or self._ptype == DOCGEN):
            raise ValueError(
                "optionclass may only be set for REPORT/TOOL/DOCGEN plugins"
            )
        self._optionclass = optionclass

    def _get_optionclass(self):
        return self._optionclass

    # TOOL attributes
    def _set_toolclass(self, toolclass):
        if not self._ptype == TOOL:
            raise ValueError("toolclass may only be set for TOOL plugins")
        self._toolclass = toolclass

    def _get_toolclass(self):
        return self._toolclass

    def _set_tool_modes(self, tool_modes):
        if not self._ptype == TOOL:
            raise ValueError("tool_modes may only be set for TOOL plugins")
        if not isinstance(tool_modes, list):
            raise ValueError("tool_modes must be a list")
        self._tool_modes = [x for x in tool_modes if x in TOOL_MODES]
        if not self._tool_modes:
            raise ValueError("tool_modes not a valid list of modes")

    def _get_tool_modes(self):
        return self._tool_modes

    require_active = property(_get_require_active, _set_require_active)
    reportclass = property(_get_reportclass, _set_reportclass)
    report_modes = property(_get_report_modes, _set_report_modes)
    category = property(_get_category, _set_category)
    optionclass = property(_get_optionclass, _set_optionclass)
    toolclass = property(_get_toolclass, _set_toolclass)
    tool_modes = property(_get_tool_modes, _set_tool_modes)

    # DOCGEN attributes
    def _set_paper(self, paper):
        if not self._ptype == DOCGEN:
            raise ValueError("paper may only be set for DOCGEN plugins")
        if not isinstance(paper, bool):
            raise ValueError("Plugin must have paper=True or False")
        self._paper = paper

    def _get_paper(self):
        return self._paper

    def _set_style(self, style):
        if not self._ptype == DOCGEN:
            raise ValueError("style may only be set for DOCGEN plugins")
        if not isinstance(style, bool):
            raise ValueError("Plugin must have style=True or False")
        self._style = style

    def _get_style(self):
        return self._style

    def _set_extension(self, extension):
        if not (
            self._ptype == DOCGEN or self._ptype == EXPORT or self._ptype == IMPORT
        ):
            raise ValueError(
                "extension may only be set for DOCGEN/EXPORT/" "IMPORT plugins"
            )
        self._extension = extension

    def _get_extension(self):
        return self._extension

    paper = property(_get_paper, _set_paper)
    style = property(_get_style, _set_style)
    extension = property(_get_extension, _set_extension)

    # QUICKREPORT attributes
    def _set_runfunc(self, runfunc):
        if not self._ptype == QUICKREPORT:
            raise ValueError("runfunc may only be set for QUICKREPORT plugins")
        self._runfunc = runfunc

    def _get_runfunc(self):
        return self._runfunc

    runfunc = property(_get_runfunc, _set_runfunc)

    # MAPSERVICE attributes
    def _set_mapservice(self, mapservice):
        if not self._ptype == MAPSERVICE:
            raise ValueError("mapservice may only be set for MAPSERVICE plugins")
        self._mapservice = mapservice

    def _get_mapservice(self):
        return self._mapservice

    mapservice = property(_get_mapservice, _set_mapservice)

    # EXPORT attributes
    def _set_export_function(self, export_function):
        if not self._ptype == EXPORT:
            raise ValueError("export_function may only be set for EXPORT plugins")
        self._export_function = export_function

    def _get_export_function(self):
        return self._export_function

    def _set_export_options(self, export_options):
        if not self._ptype == EXPORT:
            raise ValueError("export_options may only be set for EXPORT plugins")
        self._export_options = export_options

    def _get_export_options(self):
        return self._export_options

    def _set_export_options_title(self, export_options_title):
        if not self._ptype == EXPORT:
            raise ValueError("export_options_title may only be set for EXPORT plugins")
        self._export_options_title = export_options_title

    def _get_export_options_title(self):
        return self._export_options_title

    export_function = property(_get_export_function, _set_export_function)
    export_options = property(_get_export_options, _set_export_options)
    export_options_title = property(
        _get_export_options_title, _set_export_options_title
    )

    # IMPORT attributes
    def _set_import_function(self, import_function):
        if not self._ptype == IMPORT:
            raise ValueError("import_function may only be set for IMPORT plugins")
        self._import_function = import_function

    def _get_import_function(self):
        return self._import_function

    import_function = property(_get_import_function, _set_import_function)

    # GRAMPLET attributes
    def _set_gramplet(self, gramplet):
        if not self._ptype == GRAMPLET:
            raise ValueError("gramplet may only be set for GRAMPLET plugins")
        self._gramplet = gramplet

    def _get_gramplet(self):
        return self._gramplet

    def _set_height(self, height):
        if not self._ptype == GRAMPLET:
            raise ValueError("height may only be set for GRAMPLET plugins")
        if not isinstance(height, int):
            raise ValueError("Plugin must have height an integer")
        self._height = height

    def _get_height(self):
        return self._height

    def _set_detached_height(self, detached_height):
        if not self._ptype == GRAMPLET:
            raise ValueError("detached_height may only be set for GRAMPLET plugins")
        if not isinstance(detached_height, int):
            raise ValueError("Plugin must have detached_height an integer")
        self._detached_height = detached_height

    def _get_detached_height(self):
        return self._detached_height

    def _set_detached_width(self, detached_width):
        if not self._ptype == GRAMPLET:
            raise ValueError("detached_width may only be set for GRAMPLET plugins")
        if not isinstance(detached_width, int):
            raise ValueError("Plugin must have detached_width an integer")
        self._detached_width = detached_width

    def _get_detached_width(self):
        return self._detached_width

    def _set_expand(self, expand):
        if not self._ptype == GRAMPLET:
            raise ValueError("expand may only be set for GRAMPLET plugins")
        if not isinstance(expand, bool):
            raise ValueError("Plugin must have expand as a bool")
        self._expand = expand

    def _get_expand(self):
        return self._expand

    def _set_gramplet_title(self, gramplet_title):
        if not self._ptype == GRAMPLET:
            raise ValueError("gramplet_title may only be set for GRAMPLET plugins")
        if not isinstance(gramplet_title, str):
            raise ValueError(
                "gramplet_title is type %s, string or unicode required"
                % type(gramplet_title)
            )
        self._gramplet_title = gramplet_title

    def _get_gramplet_title(self):
        return self._gramplet_title

    def _set_navtypes(self, navtypes):
        if not self._ptype == GRAMPLET:
            raise ValueError("navtypes may only be set for GRAMPLET plugins")
        self._navtypes = navtypes

    def _get_navtypes(self):
        return self._navtypes

    def _set_orientation(self, orientation):
        if not self._ptype == GRAMPLET:
            raise ValueError("orientation may only be set for GRAMPLET plugins")
        self._orientation = orientation

    def _get_orientation(self):
        return self._orientation

    gramplet = property(_get_gramplet, _set_gramplet)
    height = property(_get_height, _set_height)
    detached_height = property(_get_detached_height, _set_detached_height)
    detached_width = property(_get_detached_width, _set_detached_width)
    expand = property(_get_expand, _set_expand)
    gramplet_title = property(_get_gramplet_title, _set_gramplet_title)
    navtypes = property(_get_navtypes, _set_navtypes)
    orientation = property(_get_orientation, _set_orientation)

    def _set_viewclass(self, viewclass):
        if not self._ptype == VIEW:
            raise ValueError("viewclass may only be set for VIEW plugins")
        self._viewclass = viewclass

    def _get_viewclass(self):
        return self._viewclass

    def _set_stock_category_icon(self, stock_category_icon):
        if not self._ptype == VIEW:
            raise ValueError("stock_category_icon may only be set for VIEW plugins")
        self._stock_category_icon = stock_category_icon

    def _get_stock_category_icon(self):
        return self._stock_category_icon

    def _set_stock_icon(self, stock_icon):
        if not self._ptype == VIEW:
            raise ValueError("stock_icon may only be set for VIEW plugins")
        self._stock_icon = stock_icon

    def _get_stock_icon(self):
        return self._stock_icon

    viewclass = property(_get_viewclass, _set_viewclass)
    stock_category_icon = property(_get_stock_category_icon, _set_stock_category_icon)
    stock_icon = property(_get_stock_icon, _set_stock_icon)

    # SIDEBAR attributes
    def _set_sidebarclass(self, sidebarclass):
        if not self._ptype == SIDEBAR:
            raise ValueError("sidebarclass may only be set for SIDEBAR plugins")
        self._sidebarclass = sidebarclass

    def _get_sidebarclass(self):
        return self._sidebarclass

    def _set_menu_label(self, menu_label):
        if not self._ptype == SIDEBAR:
            raise ValueError("menu_label may only be set for SIDEBAR plugins")
        self._menu_label = menu_label

    def _get_menu_label(self):
        return self._menu_label

    sidebarclass = property(_get_sidebarclass, _set_sidebarclass)
    menu_label = property(_get_menu_label, _set_menu_label)

    # VIEW, SIDEBAR and THUMBNAILER attributes
    def _set_order(self, order):
        if not self._ptype in (VIEW, SIDEBAR, THUMBNAILER):
            raise ValueError(
                "order may only be set for VIEW/SIDEBAR/THUMBNAILER plugins"
            )
        self._order = order

    def _get_order(self):
        return self._order

    order = property(_get_order, _set_order)

    # DATABASE attributes
    def _set_databaseclass(self, databaseclass):
        if not self._ptype == DATABASE:
            raise ValueError("databaseclass may only be set for DATABASE plugins")
        self._databaseclass = databaseclass

    def _get_databaseclass(self):
        return self._databaseclass

    def _set_reset_system(self, reset_system):
        if not self._ptype == DATABASE:
            raise ValueError("reset_system may only be set for DATABASE plugins")
        self._reset_system = reset_system

    def _get_reset_system(self):
        return self._reset_system

    databaseclass = property(_get_databaseclass, _set_databaseclass)
    reset_system = property(_get_reset_system, _set_reset_system)

    # GENERAL attr
    def _set_data(self, data):
        if not self._ptype in (GENERAL,):
            raise ValueError("data may only be set for GENERAL plugins")
        self._data = data

    def _get_data(self):
        return self._data

    def _set_process(self, process):
        if not self._ptype in (GENERAL,):
            raise ValueError("process may only be set for GENERAL plugins")
        self._process = process

    def _get_process(self):
        return self._process

    data = property(_get_data, _set_data)
    process = property(_get_process, _set_process)

    # RULE attr
    def _set_ruleclass(self, data):
        if self._ptype != RULE:
            raise ValueError("ruleclass may only be set for RULE plugins")
        self._ruleclass = data

    def _get_ruleclass(self):
        return self._ruleclass

    def _set_namespace(self, data):
        if self._ptype != RULE:
            raise ValueError("namespace may only be set for RULE plugins")
        self._namespace = data

    def _get_namespace(self):
        return self._namespace

    ruleclass = property(_get_ruleclass, _set_ruleclass)
    namespace = property(_get_namespace, _set_namespace)

    # THUMBNAILER attr
    def _set_thumbnailer(self, data):
        if self._ptype != THUMBNAILER:
            raise ValueError("thumbnailer may only be set for THUMBNAILER plugins")
        self._thumbnailer = data

    def _get_thumbnailer(self):
        return self._thumbnailer

    thumbnailer = property(_get_thumbnailer, _set_thumbnailer)


def newplugin():
    """
    Function to create a new plugindata object, add it to list of
    registered plugins

    :returns: a newly created PluginData which is already part of the register
    """
    gpr = PluginRegister.get_instance()
    pgd = PluginData()
    gpr.add_plugindata(pgd)
    return pgd


def register(ptype, **kwargs):
    """
    Convenience function to register a new plugin using a dictionary as input.
    The register functions will call newplugin() function, and use the
    dictionary kwargs to assign data to the PluginData newplugin() created,
    as in: plugindata.key = data

    :param ptype: the plugin type, one of REPORT, TOOL, ...
    :param kwargs: dictionary with keys attributes of the plugin, and data
                   the value
    :returns: a newly created PluginData which is already part of the register
              and which has kwargs assigned as attributes
    """
    plg = newplugin()
    plg.ptype = ptype
    for prop in kwargs:
        # check it is a valid attribute with getattr
        getattr(plg, prop)
        # set the value
        setattr(plg, prop, kwargs[prop])
    return plg


def make_environment(**kwargs):
    env = {
        "newplugin": newplugin,
        "register": register,
        "UNSTABLE": UNSTABLE,
        "EXPERIMENTAL": EXPERIMENTAL,
        "BETA": BETA,
        "STABLE": STABLE,
        "EVERYONE": EVERYONE,
        "DEVELOPER": DEVELOPER,
        "EXPERT": EXPERT,
        "REPORT": REPORT,
        "QUICKREPORT": QUICKREPORT,
        "TOOL": TOOL,
        "IMPORT": IMPORT,
        "EXPORT": EXPORT,
        "DOCGEN": DOCGEN,
        "GENERAL": GENERAL,
        "RULE": RULE,
        "MAPSERVICE": MAPSERVICE,
        "VIEW": VIEW,
        "RELCALC": RELCALC,
        "GRAMPLET": GRAMPLET,
        "SIDEBAR": SIDEBAR,
        "THUMBNAILER": THUMBNAILER,
        "CITE": CITE,
        "CATEGORY_TEXT": CATEGORY_TEXT,
        "CATEGORY_DRAW": CATEGORY_DRAW,
        "CATEGORY_CODE": CATEGORY_CODE,
        "CATEGORY_WEB": CATEGORY_WEB,
        "CATEGORY_BOOK": CATEGORY_BOOK,
        "CATEGORY_GRAPHVIZ": CATEGORY_GRAPHVIZ,
        "CATEGORY_TREE": CATEGORY_TREE,
        "TOOL_DEBUG": TOOL_DEBUG,
        "TOOL_ANAL": TOOL_ANAL,
        "TOOL_DBPROC": TOOL_DBPROC,
        "TOOL_DBFIX": TOOL_DBFIX,
        "TOOL_REVCTL": TOOL_REVCTL,
        "TOOL_UTILS": TOOL_UTILS,
        "CATEGORY_QR_MISC": CATEGORY_QR_MISC,
        "CATEGORY_QR_PERSON": CATEGORY_QR_PERSON,
        "CATEGORY_QR_FAMILY": CATEGORY_QR_FAMILY,
        "CATEGORY_QR_EVENT": CATEGORY_QR_EVENT,
        "CATEGORY_QR_SOURCE": CATEGORY_QR_SOURCE,
        "CATEGORY_QR_CITATION": CATEGORY_QR_CITATION,
        "CATEGORY_QR_SOURCE_OR_CITATION": CATEGORY_QR_SOURCE_OR_CITATION,
        "CATEGORY_QR_PLACE": CATEGORY_QR_PLACE,
        "CATEGORY_QR_MEDIA": CATEGORY_QR_MEDIA,
        "CATEGORY_QR_REPOSITORY": CATEGORY_QR_REPOSITORY,
        "CATEGORY_QR_NOTE": CATEGORY_QR_NOTE,
        "CATEGORY_QR_DATE": CATEGORY_QR_DATE,
        "REPORT_MODE_GUI": REPORT_MODE_GUI,
        "REPORT_MODE_BKI": REPORT_MODE_BKI,
        "REPORT_MODE_CLI": REPORT_MODE_CLI,
        "TOOL_MODE_GUI": TOOL_MODE_GUI,
        "TOOL_MODE_CLI": TOOL_MODE_CLI,
        "DATABASE": DATABASE,
        "GRAMPSVERSION": GRAMPSVERSION,
        "START": START,
        "END": END,
        "IMAGE_DIR": IMAGE_DIR,
    }
    env.update(kwargs)
    return env


# -------------------------------------------------------------------------
#
# PluginRegister
#
# -------------------------------------------------------------------------
class PluginRegister:
    """
    PluginRegister is a Singleton which holds plugin data

    .. attribute : stable_only
        Bool, include stable plugins only or not. Default True
    """

    __instance = None

    def get_instance():
        """Use this function to get the instance of the PluginRegister"""
        if PluginRegister.__instance is None:
            PluginRegister.__instance = 1  # Set to 1 for __init__()
            PluginRegister.__instance = PluginRegister()
        return PluginRegister.__instance

    get_instance = staticmethod(get_instance)

    def __init__(self):
        """This function should only be run once by get_instance()"""
        if PluginRegister.__instance != 1:
            raise Exception(
                "This class is a singleton. " "Use the get_instance() method"
            )
        self.stable_only = True
        if DEBUG:
            self.stable_only = False
        self.__plugindata = []
        self.__id_to_pdata = {}
        self.__req = Requirements()

    def add_plugindata(self, plugindata):
        """This is used to add an entry to the registration list.  The way it
        is used, this entry is not yet filled in, so we cannot use the id to
        add to the __id_to_pdata dict at this time."""
        self.__plugindata.append(plugindata)

    def scan_dir(self, dir, filenames, uistate=None):
        """
        The dir name will be scanned for plugin registration code, which will
        be loaded in :class:`PluginData` objects if they satisfy some checks.

        :returns: A list with :class:`PluginData` objects
        """
        # if the directory does not exist, do nothing
        if not (os.path.isdir(dir) or os.path.islink(dir)):
            return []

        ext = r".gpr.py"
        extlen = -len(ext)
        pymod = re.compile(r"^(.*)\.py$")

        for filename in filenames:
            if not filename[extlen:] == ext:
                continue
            lenpd = len(self.__plugindata)
            full_filename = os.path.join(dir, filename)
            try:
                with open(full_filename, "r", encoding="utf-8") as fd:
                    stream = fd.read()
            except Exception as msg:
                print(
                    _("ERROR: Failed reading plugin registration %(filename)s")
                    % {"filename": filename}
                )
                print(msg)
                continue
            if os.path.exists(os.path.join(os.path.dirname(full_filename), "locale")):
                try:
                    local_gettext = glocale.get_addon_translator(full_filename).gettext
                except ValueError:
                    print(
                        _(
                            "WARNING: Plugin %(plugin_name)s has no translation"
                            " for any of your configured languages, using US"
                            " English instead"
                        )
                        % {"plugin_name": filename.split(".")[0]}
                    )
                    local_gettext = glocale.translation.gettext
            else:
                local_gettext = glocale.translation.gettext
            try:
                exec(
                    compile(stream, filename, "exec"),
                    make_environment(_=local_gettext),
                    {"uistate": uistate},
                )
                for pdata in self.__plugindata[lenpd:]:
                    if pdata.id in self.__id_to_pdata:
                        # reloading
                        old = self.__id_to_pdata[pdata.id]
                        self.__plugindata.remove(old)
                        lenpd -= 1
                    self.__id_to_pdata[pdata.id] = pdata
            except ValueError as msg:
                print(
                    _("ERROR: Failed reading plugin registration %(filename)s")
                    % {"filename": filename}
                )
                print(msg)
                self.__plugindata = self.__plugindata[:lenpd]
            except:
                print(
                    _("ERROR: Failed reading plugin registration %(filename)s")
                    % {"filename": filename}
                )
                print("".join(traceback.format_exception(*sys.exc_info())))
                self.__plugindata = self.__plugindata[:lenpd]
            # check if:
            #  1. plugin exists, if not remove, otherwise set module name
            #  2. plugin not stable, if stable_only=True, remove
            #  3. TOOL_DEBUG only if DEBUG True
            rmlist = []
            ind = lenpd - 1
            for plugin in self.__plugindata[lenpd:]:
                # LOG.warning("\nPlugin scanned %s at registration", plugin.id)
                ind += 1
                plugin.directory = dir
                if not valid_plugin_version(plugin.gramps_target_version):
                    print(
                        _(
                            "ERROR: Plugin file %(filename)s has a version of "
                            '"%(gramps_target_version)s" which is invalid for Gramps '
                            '"%(gramps_version)s".'
                            % {
                                "filename": os.path.join(dir, plugin.fname),
                                "gramps_version": GRAMPSVERSION,
                                "gramps_target_version": plugin.gramps_target_version,
                            }
                        )
                    )
                    rmlist.append(ind)
                    continue
                if not self.__req.check_plugin(plugin):
                    rmlist.append(ind)
                    continue
                if plugin.status == UNSTABLE and self.stable_only:
                    rmlist.append(ind)
                    continue
                if plugin.ptype == TOOL and plugin.category == TOOL_DEBUG and not DEBUG:
                    rmlist.append(ind)
                    continue
                if plugin.fname is None:
                    continue
                match = pymod.match(plugin.fname)
                if not match:
                    rmlist.append(ind)
                    print(
                        _(
                            "ERROR: Wrong python file %(filename)s in register file "
                            "%(regfile)s"
                        )
                        % {
                            "filename": os.path.join(dir, plugin.fname),
                            "regfile": os.path.join(dir, filename),
                        }
                    )
                    continue
                if not os.path.isfile(os.path.join(dir, plugin.fname)):
                    rmlist.append(ind)
                    print(
                        _(
                            "ERROR: Python file %(filename)s in register file "
                            "%(regfile)s does not exist"
                        )
                        % {
                            "filename": os.path.join(dir, plugin.fname),
                            "regfile": os.path.join(dir, filename),
                        }
                    )
                    continue
                module = match.groups()[0]
                plugin.mod_name = module
                plugin.fpath = dir
                # LOG.warning("\nPlugin added %s at registration", plugin.id)
            rmlist.reverse()
            for ind in rmlist:
                del self.__id_to_pdata[self.__plugindata[ind].id]
                del self.__plugindata[ind]

    def get_plugin(self, id):
        """
        Return the :class:`PluginData` for the plugin with id
        """
        assert len(self.__id_to_pdata) == len(self.__plugindata)
        # if len(self.__id_to_pdata) != len(self.__plugindata):
        #     print(len(self.__id_to_pdata), len(self.__plugindata))
        return self.__id_to_pdata.get(id, None)

    def type_plugins(self, ptype):
        """
        Return a list of :class:`PluginData` that are of type ptype
        """
        return [x for x in self.__plugindata if x.ptype == ptype]

    def report_plugins(self, gui=True):
        """
        Return a list of gui or cli :class:`PluginData` that are of type REPORT

        :param gui: bool, if True then gui plugin, otherwise cli plugin
        """
        if gui:
            return [
                x
                for x in self.type_plugins(REPORT)
                if REPORT_MODE_GUI in x.report_modes
            ]
        else:
            return [
                x
                for x in self.type_plugins(REPORT)
                if REPORT_MODE_CLI in x.report_modes
            ]

    def tool_plugins(self, gui=True):
        """
        Return a list of :class:`PluginData` that are of type TOOL
        """
        if gui:
            return [x for x in self.type_plugins(TOOL) if TOOL_MODE_GUI in x.tool_modes]
        else:
            return [x for x in self.type_plugins(TOOL) if TOOL_MODE_CLI in x.tool_modes]

    def bookitem_plugins(self):
        """
        Return a list of REPORT :class:`PluginData` that are can be used as
        bookitem
        """
        return [
            x for x in self.type_plugins(REPORT) if REPORT_MODE_BKI in x.report_modes
        ]

    def quickreport_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type QUICKREPORT
        """
        return self.type_plugins(QUICKREPORT)

    def import_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type IMPORT
        """
        return self.type_plugins(IMPORT)

    def export_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type EXPORT
        """
        return self.type_plugins(EXPORT)

    def docgen_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type DOCGEN
        """
        return self.type_plugins(DOCGEN)

    def general_plugins(self, category=None):
        """
        Return a list of :class:`PluginData` that are of type GENERAL
        """
        plugins = self.type_plugins(GENERAL)
        if category:
            return [plugin for plugin in plugins if plugin.category == category]
        return plugins

    def mapservice_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type MAPSERVICE
        """
        return self.type_plugins(MAPSERVICE)

    def view_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type VIEW
        """
        return self.type_plugins(VIEW)

    def relcalc_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type RELCALC
        """
        return self.type_plugins(RELCALC)

    def gramplet_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type GRAMPLET
        """
        return self.type_plugins(GRAMPLET)

    def sidebar_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type SIDEBAR
        """
        return self.type_plugins(SIDEBAR)

    def database_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type DATABASE
        """
        return self.type_plugins(DATABASE)

    def rule_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type RULE
        """
        return self.type_plugins(RULE)

    def thumbnailer_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type THUMBNAILER
        """
        return self.type_plugins(THUMBNAILER)

    def cite_plugins(self):
        """
        Return a list of :class:`PluginData` that are of type CITE
        """
        return self.type_plugins(CITE)

    def filter_load_on_reg(self):
        """
        Return a list of :class:`PluginData` that have load_on_reg == True
        """
        return [x for x in self.__plugindata if x.load_on_reg == True]
